package org.expeditee.gui;

import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import org.expeditee.io.Conversion;
import org.expeditee.items.Item;
import org.expeditee.items.Text;

/**
 * This class provides the methods to extract and set attributes of Items and
 * Frames. These methods are called when a user merges a text item with
 * <code>Attribute: Value</code> pairs.
 * 
 * @author jdm18
 * 
 */
public class AttributeUtils {

	public static final char SEPARATOR_CHAR = ':';

	private static final String SEPARATOR_STRING = SEPARATOR_CHAR + " ";

	private static final int GET_LENGTH = "get".length();

	private static int SET_LENGTH = "set".length();

	// List of method names to ignore when searching for a match
	private static List<Method> _GetMethods = null;

	private static HashMap<String, Method> _SetMethods = null;

	// List of attributes which are ignored when copying.
	private static List<String> _Ignore = null;

	// List of method names to show in extraced lists even when they return
	// null
	// (Null is often used to indicate the default value is used)
	private static List<Method> _AllowNull = null;

	private static List<String> _ExtractIgnore = null;

	// private static HashMap<String, String> _Abbreviations = null;

	/**
	 * Initialises the _Ignore and _AllowNull lists.
	 */
	private static void initLists() {

		Class[] param = {};

		try {
			_Ignore = new LinkedList<String>();
			_Ignore.add("date");
			_Ignore.add("datecreated");
			_Ignore.add("d");
			_Ignore.add("link");
			_Ignore.add("l");
			_Ignore.add("action");
			_Ignore.add("a");
			_Ignore.add("position");
			_Ignore.add("pos");
			_Ignore.add("p");
			_Ignore.add("x");
			_Ignore.add("y");

			_ExtractIgnore = new LinkedList<String>();
			_ExtractIgnore.add("x");
			_ExtractIgnore.add("y");

			_AllowNull = new LinkedList<Method>();
			_AllowNull.add(Item.class.getMethod("getColor", param));
			_AllowNull.add(Item.class.getMethod("getBackgroundColor", param));

			_AllowNull.add(Frame.class.getMethod("getBackgroundColor", param));
			_AllowNull.add(Frame.class.getMethod("getForegroundColor", param));

			_GetMethods = new LinkedList<Method>();
			_GetMethods.add(Item.class.getMethod("getDateCreated", param));

			_GetMethods.add(Item.class.getMethod("getColor", param));
			_GetMethods.add(Item.class.getMethod("getBackgroundColor", param));
			_GetMethods.add(Item.class.getMethod("getAction", param));
			_GetMethods.add(Item.class.getMethod("getData", param));
			_GetMethods.add(Item.class.getMethod("getLink", param));
			_GetMethods.add(Item.class.getMethod("getFillColor", param));
			_GetMethods.add(Item.class.getMethod("getFillPattern", param));
			_GetMethods.add(Item.class.getMethod("getThickness", param));

			_GetMethods.add(Item.class.getMethod("getOwner", param));
			_GetMethods.add(Item.class.getMethod("getLinkMark", param));
			_GetMethods.add(Item.class.getMethod("getActionMark", param));

			_GetMethods
					.add(Item.class.getMethod("getActionCursorEnter", param));
			_GetMethods
					.add(Item.class.getMethod("getActionCursorLeave", param));
			_GetMethods.add(Item.class.getMethod("getActionEnterFrame", param));
			_GetMethods.add(Item.class.getMethod("getActionLeaveFrame", param));

			_GetMethods.add(Item.class.getMethod("getTopShadowColor", param));
			_GetMethods
					.add(Item.class.getMethod("getBottomShadowColor", param));
			_GetMethods.add(Item.class.getMethod("getArrow", param));

			_GetMethods.add(Item.class.getMethod("getLinePattern", param));
			_GetMethods.add(Item.class.getMethod("getLinkFrameset", param));
			_GetMethods.add(Item.class.getMethod("getLinkTemplate", param));
			_GetMethods.add(Item.class.getMethod("getPosition", param));
			_GetMethods.add(Item.class.getMethod("getX", param));
			_GetMethods.add(Item.class.getMethod("getY", param));

			_GetMethods.add(Text.class.getMethod("getFamily", param));
			_GetMethods.add(Text.class.getMethod("getFontStyle", param));
			_GetMethods.add(Text.class.getMethod("getJustification", param));
			_GetMethods.add(Text.class.getMethod("getWidth", param));
			_GetMethods.add(Item.class.getMethod("getSize", param));

			_GetMethods.add(Frame.class.getMethod("getOwner", param));
			_GetMethods.add(Frame.class.getMethod("getProtection", param));
			_GetMethods.add(Frame.class.getMethod("getDateCreated", param));
			_GetMethods.add(Frame.class.getMethod("getLastModifyUser", param));
			_GetMethods.add(Frame.class.getMethod("getLastModifyDate", param));
			_GetMethods.add(Frame.class.getMethod("getForegroundColor", param));
			_GetMethods.add(Frame.class.getMethod("getBackgroundColor", param));

			Class[] pPoint = { Point.class };
			Class[] pString = { String.class };
			Class[] pInt = { int.class };
			Class[] pFloat = { float.class };
			Class[] pColor = { Color.class };
			Class[] pBool = { boolean.class };
			Class[] pArrow = { int.class, double.class };
			Class[] pList = { List.class };
			Class[] pIntArray = { int[].class };

			_SetMethods = new HashMap<String, Method>();
			_SetMethods.put("x", Item.class.getMethod("setX", pInt));
			_SetMethods.put("y", Item.class.getMethod("setY", pInt));
			_SetMethods.put("position", Item.class.getMethod("setPosition",
					pPoint));
			_SetMethods.put("p", Item.class.getMethod("setPosition", pPoint));
			_SetMethods.put("pos", Item.class.getMethod("setPosition", pPoint));
			_SetMethods.put("thickness", Item.class.getMethod("setThickness",
					pFloat));
			_SetMethods.put("t", Item.class.getMethod("setThickness", pFloat));

			_SetMethods.put("color", Item.class.getMethod("setColor", pColor));
			_SetMethods.put("c", Item.class.getMethod("setColor", pColor));

			_SetMethods.put("backgroundcolor", Item.class.getMethod(
					"setBackgroundColor", pColor));
			_SetMethods.put("bgc", Item.class.getMethod("setBackgroundColor",
					pColor));

			_SetMethods
					.put("action", Item.class.getMethod("setActions", pList));
			_SetMethods.put("a", Item.class.getMethod("setActions", pList));
			_SetMethods.put("d", Item.class.getMethod("setData", pList));
			_SetMethods.put("data", Item.class.getMethod("setData", pList));

			_SetMethods.put("link", Item.class.getMethod("setLink", pString));
			_SetMethods.put("l", Item.class.getMethod("setLink", pString));

			_SetMethods.put("fillcolor", Item.class.getMethod("setFillColor",
					pColor));
			_SetMethods.put("fc", Item.class.getMethod("setFillColor", pColor));

			_SetMethods.put("fillpattern", Item.class.getMethod(
					"setFillPattern", pString));
			_SetMethods.put("fp", Item.class.getMethod("setFillPattern",
					pString));

			_SetMethods.put("owner", Item.class.getMethod("setOwner", pString));
			_SetMethods.put("linkmark", Item.class.getMethod("setLinkMark",
					pBool));
			_SetMethods.put("lm", Item.class.getMethod("setLinkMark", pBool));
			_SetMethods.put("actionmark", Item.class.getMethod("setActionMark",
					pBool));
			_SetMethods.put("am", Item.class.getMethod("setActionMark", pBool));

			_SetMethods.put("actioncursorenter", Item.class.getMethod(
					"setActionCursorEnter", pList));
			_SetMethods.put("actioncursorleave", Item.class.getMethod(
					"setActionCursorLeave", pList));
			_SetMethods.put("actionenterframe", Item.class.getMethod(
					"setActionEnterFrame", pList));
			_SetMethods.put("actionleaveframe", Item.class.getMethod(
					"setActionLeaveFrame", pList));

			_SetMethods.put("topshadow", Item.class.getMethod(
					"setTopShadowColor", pColor));
			_SetMethods.put("bottomshadow", Item.class.getMethod(
					"setBottomShadowColor", pColor));
			_SetMethods.put("arrow", Item.class.getMethod("setArrow", pArrow));

			_SetMethods.put("linepattern", Item.class.getMethod(
					"setLinePattern", pIntArray));
			_SetMethods.put("lp", Item.class.getMethod("setLinePattern",
					pIntArray));

			_SetMethods.put("linkframeset", Item.class.getMethod(
					"setLinkFrameset", pString));
			_SetMethods.put("lf", Item.class.getMethod("setLinkFrameset",
					pString));
			_SetMethods.put("linktemplate", Item.class.getMethod(
					"setLinkTemplate", pString));
			_SetMethods.put("lt", Item.class.getMethod("setLinkTemplate",
					pString));

			_SetMethods.put("family", Text.class
					.getMethod("setFamily", pString));
			_SetMethods.put("face", Text.class.getMethod("setFontStyle",
					pString));
			_SetMethods.put("fontstyle", Text.class.getMethod("setFontStyle",
					pString));
			_SetMethods.put("justification", Text.class.getMethod(
					"setJustification", pInt));
			_SetMethods.put("width", Text.class.getMethod("setWidth", pInt));
			_SetMethods.put("size", Item.class.getMethod("setSize", pInt));
			_SetMethods.put("s", Item.class.getMethod("setSize", pInt));

			_SetMethods.put("foregroundcolor", Frame.class.getMethod(
					"setForegroundColor", pColor));
			_SetMethods.put("fgc", Frame.class.getMethod("setForegroundColor",
					pColor));
			_SetMethods.put("backgroundcolor0", Frame.class.getMethod(
					"setBackgroundColor", pColor));
			_SetMethods.put("bgc0", Frame.class.getMethod("setBackgroundColor",
					pColor));
			_SetMethods.put("protection", Frame.class.getMethod(
					"setProtection", pString));

		} catch (SecurityException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Extracts a list of attributes from the given Object. Any method that
	 * starts with <code>get</code>, takes no arguments and is not found in
	 * the Ignore list will be run, All the attributes are then put into a Text
	 * Item of the form <Name>:<Value> If the value returned by the get method
	 * is null, then the attribute will not be included, unless the name of the
	 * method is found in the AllowNull list.
	 * 
	 * @param toExtract
	 *            The Object from which to extract the attributes
	 * @return A Text Item containing the extracted Attributes.
	 */
	public static Item ExtractAttributes(Object toExtract) {
		if (toExtract == null)
			return null;

		// ensure the lists are populated
		if (_Ignore == null)
			initLists();

		// StringBuffer to store all the extracted Attribute:Value pairs
		StringBuffer attributes = new StringBuffer();

		// iterate through the list of methods
		for (Method m : _GetMethods) {

			// Make sure the classes of the methods match the item
			if (m.getDeclaringClass().isAssignableFrom(toExtract.getClass())) {
				try {
					Object o = m.invoke(toExtract, (Object[]) null);

					if (o == null) {
						// methods that return null are only included if they
						// are in the AllowNull list
						if (_AllowNull.contains(m)) {
							String name = m.getName().substring(
									GET_LENGTH).toLowerCase();
							if (name.equals("color"))
								o = "default";
							else if (name.equals("backgroundcolor"))
								o = "transparent";
							else if (name.equals("foregroundcolor"))
								o = "auto";
							else
								o = "";
						} else {
							continue;
						}
					}
					// skip methods that are in the ignore lists
					if (_ExtractIgnore.contains(m.getName().substring(
							GET_LENGTH).toLowerCase())) {
						continue;
					}

					if (o instanceof Integer) {
						Integer i =  (Integer)o;
						if (i == Item.DEFAULT_INTEGER)
							continue;
						if (m.getName().endsWith("Justification")
								&& convertJustificationToString((Integer) o) != null)
							o = convertJustificationToString((Integer) o);
						// -1 indicates default value
						else 
							o = i;
					} else if (o instanceof Float) {
						// -1 indicates default value
						if (((Float) o) < 0.0001)
							continue;
					} else if (o instanceof Double) {
						// -1 indicates default value
						if (((Double) o) < 0.0001)
							continue;
					} else if (o instanceof Color) {
						// converts the color to the Expeditee code
						o = Conversion.getExpediteeColorCode((Color) o);
						if (o == null)
							continue;
					} else if (o instanceof Point) {
						Point p = (Point) o;
						o = (int) p.getX() + " " + (int) p.getY();
					} else if (o instanceof Font) {
						Font f = (Font) o;

						String s = f.getName() + "-";
						if (f.isPlain())
							s += "Plain";

						if (f.isBold())
							s += "Bold";

						if (f.isItalic())
							s += "Italic";

						s += "-" + f.getSize();
						o = s;
					} else if (o instanceof Text) {
						o = ((Text) o).getFirstLine();
					} else if (o instanceof List) {
						List list = (List) o;
						for (Object ob : list)
							attributes
									.append(m.getName().substring(GET_LENGTH))
									.append(SEPARATOR_STRING).append(ob)
									.append("\n");
						continue;
					} else if (o instanceof int[]) {
						StringBuffer sb = new StringBuffer();
						int[] values = (int[]) o;
						for (int i = 0; i < values.length; i++) {
							sb.append(values[i]).append(' ');
						}
						sb.deleteCharAt(attributes.length() - 1);
						o = sb.toString();
					} else if (o instanceof Boolean) {
						// true is the default for boolean values
						if (((Boolean) o).booleanValue())
							continue;
					}
					//Append the attributes
					attributes.append(m.getName().substring(GET_LENGTH))
							.append(SEPARATOR_STRING).append(o).append("\n");
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}

		// if no attributes were extracted
		if (attributes.length() <= 0)
			return null;

		while (attributes.charAt(attributes.length() - 1) == '\n')
			attributes.delete(attributes.length() - 1, attributes.length());

		// create the text Item
		Frame current = DisplayIO.getCurrentFrame();
		Item attribs = current.getStatsTextItem(attributes.toString());
		return attribs;
	}

	/**
	 * Returns the String name of the justification value given. just should
	 * correspond to one of the Item.JUSTIFICATION constants, if it does not
	 * then null is returned.
	 * 
	 * @param just
	 *            The justification level as defined by the Item.JUSTIFICATION
	 *            constants.
	 * @return The String name (i.e. "Center") of the justification given, or
	 *         null if no match is found.
	 */
	private static String convertJustificationToString(int just) {
		switch (just) {
		case Item.JUSTIFICATION_CENTER:
			return "Center";
		case Item.JUSTIFICATION_FULL:
			return "Full";
		case Item.JUSTIFICATION_LEFT:
			return "Left";
		case Item.JUSTIFICATION_RIGHT:
			return "Right";
		default:
			return null;
		}
	}

	/**
	 * Attempts to set the attribute in the given attribute: value pair. The
	 * value string should be formatted as follows:
	 * <code> Attribute: Value </code> Multiple values can be used if they are
	 * separated by spaces
	 * 
	 * @param toSet
	 *            The Item to set the attribute of
	 * @param attribs
	 *            The Text item that contains the list of attributes to set
	 * @return True if the attribute(s) were sucessfully set, false otherwise
	 */
	public static boolean SetAttribute(Object toSet, Text attribs) {
		// error checking
		if (toSet == null || attribs == null)
			return false;

		if (_Ignore == null)
			initLists();

		// get the list of attribute: value pairs
		List<String> values = attribs.getTextList();
		// if no paris exist, we are done
		if (values.size() == 0
				|| (values.size() == 1 && values.get(0).length() == 0)) {
			return false;
		}

		// loop through all attribute: value pairs
		for (int i = 0; i < values.size(); i++) {
			StringBuffer v = new StringBuffer(values.get(i));

			// remove the annotation mark (if applicable)
			if (v.indexOf("@") == 0)
				v = v.deleteCharAt(0);

			// check if the next string is another attribute to merge or a
			// continuation
			while (i < values.size() - 1) {
				StringBuffer next = new StringBuffer(values.get(i + 1));

				// if the next String has a colon, then it may be another
				// attribute
				if (next.indexOf("" + SEPARATOR_CHAR) >= 0) {
					// if the attribute is the same as v, then it is a
					// continuation
					if (v.indexOf(getAttribute(next.toString())) == 0) {
						// strip the attribute from next
						next = new StringBuffer(getValue(next.toString()));
						// if the attribute is not the same, then it may be a
						// new method
					} else {
						// if this is indeed a method, then leave the while loop
						// if(_SetMethods.containsKey(StripFromColon(next.toString()).toLowerCase())){
						break;
						// }
					}
				}

				v.append("\n").append(next);
				i++;
			}

			if (v.length() > 0
					&& !SetAttribute(toSet, v.toString(), values.size() > 1)) {
				// if no other attributes have been set
				if (i == 0)
					return false;
				// otherwise, this is a list of attributes, so continue
				else {
					String stripped = getAttribute(v.toString());
					if (stripped == null) {
						// This happens when there is an attribute at the start
						// Then a bunch of plain text
						return false;
					} else if (_Ignore.contains(stripped)) {
						return false;
					} else if (!(_SetMethods.containsKey(stripped))) {
						// Display an error message if its not in our list of
						// attributes to ignore when copying
						FrameGraphics.WarningMessage("Attribute: '"
								+ getAttribute(v.toString())
								+ "' does not exist.");
					} else {
						String types = "";
						for (Class c : _SetMethods.get(stripped)
								.getParameterTypes())
							types += c.getSimpleName() + " ";
						FrameGraphics.WarningMessage("Wrong arguments for: '"
								+ getAttribute(v.toString()) + "' expecting "
								+ types.trim() + " found '"
								+ getValue(v.toString()) + "'");
					}
				}
			} else if (v.length() == 0)
				return false;
		}

		return true;
	}

	private static boolean SetAttribute(Object toSet, String value,
			boolean isAttributeList) {
		// separate attribute and value from string
		String attribute = getAttribute(value);
		value = getValue(value);

		// if there was no colon
		if (value == null || attribute == null)
			return false;

		if (value.length() == 0)
			value = "";

		attribute = attribute.toLowerCase();

		// Some properties are ignored when multiple attributes are being set on
		// an item at the same time
		if (isAttributeList && _Ignore.contains(attribute)) {
			// System.out.println("Attribute ignored: " + attribute);
			return true;
		}

		// separate multiple values if required
		Method toRun = _SetMethods.get(attribute);

		// if this is not the name of a method, it may be the name of an agent
		if (toRun == null) {
			// System.out.println("Attrib not found for: " + attribute);
			return false;
		}

		// if there are duplicate methods with the same name
		List<Method> possibles = new LinkedList<Method>();
		if (toRun.getDeclaringClass().isInstance(toSet))
			possibles.add(toRun);
		int i = 0;
		while (_SetMethods.containsKey(attribute + i)) {
			if (_SetMethods.get(attribute + i).getDeclaringClass().isAssignableFrom(toSet.getClass()))
				possibles.add(_SetMethods.get(attribute + i));
			i++;
		}

		for (Method possible : possibles) {
			Object current = null;
			Object[] param = {};
			// find the corresponding get method for this set method
			// and get the current value of the attribute
			for (Method m : _GetMethods) {
				if (m.getDeclaringClass().isAssignableFrom(toSet.getClass())
						&& m.getName().substring(GET_LENGTH).equals(
								possible.getName().substring(SET_LENGTH))) {
					try {
						current = m.invoke(toSet, param);
					} catch (IllegalArgumentException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					} catch (IllegalAccessException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					} catch (InvocationTargetException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					break;
				}

			}

			try {
				Object[] params = Conversion.Convert(possible, value, current);

				try {
					possible.invoke(toSet, params);
					return true;
				} catch (IllegalArgumentException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (InvocationTargetException e) {
					FrameGraphics.DisplayMessage(toSet.getClass()
							.getSimpleName()
							+ " type does not support that attribute.");
					// e.printStackTrace();
				}
			} catch (NumberFormatException e) {

			}
		}

		return false;

	}

	/**
	 * Returns the part of the given string that is after the attribute value
	 * pair separator. If that character is not there it returns the empty
	 * string if it is an annotation item or the entire string if it is not.
	 * 
	 * @param attributeValuePair
	 *            the string to get the value from.
	 * @return the value from the attribute value pair.
	 */
	public static String getValue(String toStrip) {
		assert (toStrip != null);
		if (toStrip.length() == 0)
			return "";

		int lineSeparator = toStrip.indexOf(Character.LINE_SEPARATOR);
		if(lineSeparator > 0){
			toStrip = toStrip.substring(0,lineSeparator);
		}
		
		int ind = toStrip.lastIndexOf(SEPARATOR_CHAR);

		if (ind < 0) {
			// If it is an annotation item return the empty string
			// Annotation items can not be values only
			if (toStrip.charAt(0) == '@') {
				return "";
			}
			// It is a value with no attribute.
			return toStrip.trim();
		}

		return toStrip.substring(ind + 1).trim();
	}

	/**
	 * Returns the part of the given string that is before the attribute value
	 * pair separator, or null if the given String does not include the
	 * separator.
	 * 
	 * @param attributeValuePair
	 *            The String to strip
	 * @return the attribute if there is one or null if there is not
	 */
	public static String getAttribute(String attributeValuePair) {
		int ind = attributeValuePair.indexOf(SEPARATOR_CHAR);
		if (ind < 0)
			return null;

		attributeValuePair = attributeValuePair.substring(0, ind);
		attributeValuePair = attributeValuePair.trim();

		if (attributeValuePair.length() == 0)
			return null;

		return attributeValuePair;
	}

	/**
	 * 
	 * @param text
	 * @param value
	 */
	public static void setSingleValue(Text text, String value) {
		assert (value != null);

		String oldText = text.getText();
		String attribute = getAttribute(oldText);

		if (attribute == null)
			attribute = text.getText().trim();

		text.setText(attribute + SEPARATOR_STRING + value);
	}

	public static Double getDoubleValue(String attributeValuePair) {
		String value = getValue(attributeValuePair);

		if (value == null)
			return null;

		try {
			return Double.parseDouble(value);
		} catch (Exception e) {
		}
		return null;
	}
}
