package org.expeditee.gui;

import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javax.swing.Timer;

import org.expeditee.actions.Actions;
import org.expeditee.actions.NavigationActions;
import org.expeditee.items.Constraint;
import org.expeditee.items.Dot;
import org.expeditee.items.InteractiveWidgetInitialisationFailedException;
import org.expeditee.items.InteractiveWidgetNotAvailableException;
import org.expeditee.items.Item;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Line;
import org.expeditee.items.Picture;
import org.expeditee.items.Text;
import org.expeditee.items.WidgetCorner;
import org.expeditee.items.WidgetEdge;
import org.expeditee.stats.SessionStats;

public class FrameMouseActions implements MouseListener, MouseMotionListener,
		MouseWheelListener {

	private static int _lastMouseClickModifiers = 0;

	private FrameMouseActions() {
	}

	private static FrameMouseActions _instance = null;

	public static FrameMouseActions getInstance() {
		if (_instance == null)
			_instance = new FrameMouseActions();
		return _instance;
	}

	private static final int RECTANGLE_CORNERS = 4;

	// TODO say where/how used
	private static final int MOUSE_WHEEL_THRESHOLD = 3;

	private static final int MINIMUM_RANGE_DEPRESS_TIME = 250;

	private static final int RECTANGLE_TO_POINT_THRESHOLD = 20;

	private static Date _lastMouseClickDate = new Date();

	public static final int LITTLE_MOUSE_PAUSE = 500;

	public static final int ZERO_MOUSE_PAUSE = 0;

	public static final int BIG_MOUSE_PAUSE = 750;

	public static final int CONTEXT_FREESPACE = 0;

	public static final int CONTEXT_AT_TEXT = 1;

	public static final int CONTEXT_AT_LINE = 2;

	public static final int CONTEXT_AT_DOT = 3;

	public static final int CONTEXT_AT_ENCLOSURE = 4;

	public static int _alpha = -1;

	/**
	 * The last known mouse X coordinate
	 */
	public static int MouseX;

	/**
	 * The last known mouse Y coordinate. Relative to the top of the
	 * application.
	 */
	public static int MouseY;

	// Distance of mouse cursor from the origin of the item that was picked up
	// The are used in the move method to calculate the distance moved by the
	// cursor
	private static int _offX;

	private static int _offY;

	// Keeps track of mouse button events when a delete occurs
	private static boolean _isDelete = false;

	// Keeps track of mouse button events when the user extracts attributes
	// occurs
	private static boolean _isAttribute = false;

	/**
	 * A flag to indicate that the last mouseUp event was part of a two button
	 * click sequence hence the next mouse up should be ignored*
	 */
	private static boolean _wasDouble = false;

	private static boolean _isNoOp = false;

	private static boolean _extrude = false;

	// keeps track of the last highlighted Item
	private static Item _lastHighlightedItem = null;

	// keeps track of the item being 'ranged out' if there is one.
	private static Text _lastRanged = null;

	// keeps track of the picture being cropped if there is one
	private static Picture _lastCropped = null;

	// true if lastItem only has highlighting removed when a new item is
	// highlighted
	private static boolean _lastHoldsHighlight = false;

	private static boolean _forceArrowCursor = true;

	// the current context of the cursor
	private static int _context = 0;

	public static void setForceArrow(boolean val) {
		_forceArrowCursor = val;
	}

	public static int getContext() {
		return _context;
	}

	static int _mouseDown = 0;

	private static MouseEvent _lastMouseClick = null;

	private static Item _lastClickedOn = null;

	private static List<Item> _lastClickedIn = null;

	private static boolean _pulseOn = false;

	private static final int PULSE_AMOUNT = 2;

	private static Timer _MouseTimer = new Timer(LITTLE_MOUSE_PAUSE,
			new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					// check if we are in free space
					if (_lastClickedOn == null && Frame.FreeItems.size() == 0) {
						// System.out.println("SuperBack!");
						_MouseTimer.setDelay(ZERO_MOUSE_PAUSE);
						back();
					} else {
						if (FrameUtils.getCurrentItem() == null) {
							// Check if we are toggling arrowhead
							if (Frame.FreeItems.size() <= 2) {
								for (Item i : Frame.FreeItems) {
									if (i instanceof Line) {
										((Line) i).toggleArrow();
									}
								}
								FrameGraphics.Repaint();
							}
						}
						_MouseTimer.stop();
					}
				}
			});

	private static void setPulse(boolean pulseOn) {
		if (_pulseOn == pulseOn)
			return;
		int amount = PULSE_AMOUNT;
		if (!pulseOn)
			amount *= -1;
		_pulseOn = pulseOn;

		for (Item i : _lastClickedOn.getAllConnected()) {
			if (i instanceof Line) {
				Line line = (Line) i;
				line.setThickness(line.getThickness() + amount);
			}
		}
		FrameGraphics.Repaint();
	}

	private static Timer _ExtrudeMouseTimer = new Timer(BIG_MOUSE_PAUSE,
			new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					setPulse(true);
					_extrude = true;
					_ExtrudeMouseTimer.stop();
				}
			});

	public void mouseClicked(MouseEvent e) {
	}

	/**
	 * Each Item on the Frame is checked to determine if the mouse x,y
	 * coordinates are on the Item (or within the Shape surrounding it). If the
	 * coordinates are on the Item then the Item is checked for a link, if it
	 * has a link the link is followed, if not, nothing is done.
	 */
	public void mousePressed(MouseEvent e) {
		ProccessMousePressedEvent(e, e.getModifiersEx());
	}

	public void ProccessMousePressedEvent(MouseEvent e, int modifiersEx) {
		// System.out.println("MousePressed " + e.getX() + "," + e.getY() + " "
		// + e.getWhen());

		// TODO WHY DID I NOT COMMENT THIS LINE!! MIKE SAYS
		if (LastRobotX != null) {
			_RobotTimer.stop();
			LastRobotX = null;
			LastRobotY = null;
			mouseMoved(e);
		}

		NavigationActions.ResetLastAddToBack();
		// System.out.println(modifiersEx);
		int buttonPressed = e.getButton();
		_mouseDown += buttonPressed;

		_lastMouseClickDate = new Date();
		_lastClickedOn = FrameUtils.getCurrentItem();
		// load any frame if necessary
		Item on = _lastClickedOn;

		_lastClickedIn = FrameUtils.getCurrentItems(on);
		// if (_lastClickedIn != null){
		// System.out.println(_lastClickedIn.size());}

		SessionStats.MouseClicked(e.getButton());
		if (buttonPressed == MouseEvent.BUTTON1) {
			SessionStats.AddFrameEvent("Ld");
			_extrude = false;
		} else if (buttonPressed == MouseEvent.BUTTON2) {
			SessionStats.AddFrameEvent("Md");
			_extrude = false;
		} else if (buttonPressed == MouseEvent.BUTTON3) {
			SessionStats.AddFrameEvent("Rd");
		}

		// Mike says...
		// For somereason the modifiers for e are different from modifiersEx
		// The SwingUtilities.convertMouseEvent method changes the modifiers
		_lastMouseClick = e;
		_lastMouseClickModifiers = modifiersEx;

		// Only start the timer when in free space when the user double clicks
		// to do super back
		// TODO change this so that there are separate timers for super back and
		// the other
		// Long depress actions if that is what is wanted.
		if (_lastClickedOn == null && Frame.FreeItems.size() == 0) {
			// System.out.println(e.getClickCount());
			if (e.getClickCount() >= 2) {
				_MouseTimer.start();
			}
		} else if (_lastClickedOn != null && Frame.FreeItems.size() == 0
				&& e.getButton() == MouseEvent.BUTTON3) {
			_ExtrudeMouseTimer.start();

		} else {
			_MouseTimer.start();
		}

		// pre-cache the frame if it is linked
		if (on != null && on.getLink() != null && on.isLinkValid()) {
			FrameIO.Precache(on.getAbsoluteLink());
		}

		// check for delete command
		if (isDelete(modifiersEx)) {
			_isDelete = true;
			// _lastRanged = null;
			_lastCropped = null;
			_wasDouble = false;
			// check for attributes command
		} else if (isGetAttributes(modifiersEx)) {
			_isAttribute = true;
			_wasDouble = false;
		} else if (isTwoClickNoOp(modifiersEx)) {
			_isAttribute = false;
			_wasDouble = false;
			_isDelete = false;
			_isNoOp = true;
		} else
			_isDelete = false;

		// This must happen before the previous code
		// This is when the user is anchoring something
		if (buttonPressed != MouseEvent.BUTTON1
				&& (_context == CONTEXT_FREESPACE || _context == CONTEXT_AT_ENCLOSURE)
				&& Frame.itemAttachedToCursor()) {
			FrameGraphics.ChangeSelectionMode(_lastHighlightedItem,
					Item.SelectedMode.None);

			_lastHighlightedItem = Frame.getItemAttachedToCursor();
			for (Item i : Frame.FreeItems) {
				i.restoreLastMode(Item.DEPRESSED_HIGHLIGHT);
			}
			FrameGraphics.Repaint();
			// this is when the user is picking something up
		} else if (_lastHighlightedItem != null) {
			if (!(_lastHighlightedItem instanceof Line)) {
				_lastHighlightedItem
						.setSelectionColor(Item.DEPRESSED_HIGHLIGHT);
			} else {
				for (Item i : _lastHighlightedItem.getAllConnected()) {
					i.setSelectionColor(Item.DEPRESSED_HIGHLIGHT);
				}
			}
			FrameGraphics.Repaint();
		}

		// if the user is ranging text
		if (on != null && on instanceof Text && !_isDelete) {
			_lastRanged = (Text) on;
			// set start-drag point
			_lastRanged.setSelectionStart(DisplayIO.getMouseX(), DisplayIO
					.getMouseY());
		}

		if (on != null && on instanceof Picture
				&& e.getButton() == MouseEvent.BUTTON3 && !_isDelete) {
			_lastCropped = (Picture) on;
			// set start crop point
			_lastCropped.setStartCrop(DisplayIO.getMouseX(), DisplayIO
					.getMouseY());
			_lastCropped.setShowCrop(true);
		}
	}

	// This is where all the processing happens
	public void mouseReleased(MouseEvent e) {
		// System.out.println("Released " + e.getX() + "," + e.getY() + " " +
		// e.getWhen());
		FrameUtils.ResponseTimer.restart();

		_mouseDown -= e.getButton();
		updateCursor();

		// System.out.println(e.getX() + ", " + e.getY());

		Text lastRanged = _lastRanged;
		_lastRanged = null;
		// Dont do ranging if the user moves really quickly...
		// They are probably trying to pick something up in this case
		if (lastRanged != null) {
			long depressTime = (new Date()).getTime()
					- _lastMouseClickDate.getTime();
			// double changeInDistance =
			// e.getPoint().distance(_currentMouseClick.getPoint());
			// double speed = changeInDistance * 1000 / changeInTime;

			// System.out.println(depressTime);

			if (depressTime < MINIMUM_RANGE_DEPRESS_TIME
					|| lastRanged.getSelectedSize() <= 0) {// Text.MINIMUM_RANGED_CHARS)
				// {
				lastRanged.clearSelection();
				lastRanged = null;
			}
		}

		_ExtrudeMouseTimer.stop();
		_MouseTimer.stop();
		setPulse(false);

		// if the last action was a delete, then ignore the next mouseup
		if (_wasDouble) {
			_wasDouble = false;
			return;
		}

		// This code must come after the _wasDouble code...
		// Otherwise get Stopping Agent method after doing the left+right format
		// shortcut
		if (Actions.isAgentRunning()) {
			Actions.stopAgent();
			return;
		}

		/*
		 * if (_isNoOp) { if (e.getButton() != MouseEvent.NOBUTTON) { _isNoOp =
		 * false; _wasDouble = true; // lastRanged.clearSelection();
		 * FrameGraphics.Repaint(); return; } }
		 */

		// get whatever the user was pointing at
		Item clickedOn = _lastClickedOn;
		List<Item> clickedIn = _lastClickedIn;

		MouseX = e.getX();
		MouseY = e.getY();

		Item releasedOn = FrameUtils.getCurrentItem();
		List<Item> releasedIn = FrameUtils.getCurrentItems(releasedOn);

		// TODO Simplify the login below
		if (_isNoOp
				&& ((clickedOn != null && (releasedIn != null || (releasedOn != null && releasedOn != clickedOn))) || (releasedOn == null && releasedIn == null))) {
			if (_isDelete) {
				_isDelete = false;
				_wasDouble = true;
			}

			_isNoOp = false;

			if (_lastHighlightedItem != null)
				FrameGraphics.ChangeSelectionMode(_lastHighlightedItem,
						Item.SelectedMode.None);

			if (Frame.itemAttachedToCursor()) {
				move(Frame.FreeItems);
			}

			FrameGraphics
					.DisplayMessage("Action cancelled, mouse moved more than "
							+ UserSettings.NoOpThreshold + " pixels.");
			FrameGraphics.Repaint();
			return;
		} else
			_isNoOp = false;

		// if this is a delete command
		if (_isDelete) {
			if (lastRanged != null) {

				Item i = Frame.getItemAttachedToCursor();
				if (i != null && i instanceof Text) {
					lastRanged.replaceSelectedText(((Text) i).getText());
					Frame.FreeItems.clear();
				} else
					lastRanged.cutSelectedText();
				lastRanged.clearSelection();
				FrameGraphics.Repaint();

			} else {
				delete(clickedOn);
			}
			_wasDouble = true;
			_isDelete = false;
			return;
		}

		// if this is an attribute extraction command
		if (_isAttribute) {
			if (clickedOn == null) {
				// First look for the @NoFormat tag
				Frame current = DisplayIO.getCurrentFrame();
				Actions.PerformAction(current, null, "Format");
				/*
				 * if (ItemUtils.ContainsTag(current.getItems(), "@NoFormat")) {
				 * FrameGraphics .DisplayMessage("Frame is protected by
				 * @NoFormat tag."); } else { new Format().process(current); }
				 */
			} else {
				extractAttributes(clickedOn);
			}
			// if the user dragged and displayed some cropping with left and
			// right button is a no op for now
			// but later could make this the shrinkTo context
			if (_lastCropped != null) {
				_lastCropped.clearCropping();
				_lastCropped = null;
			}
			_wasDouble = true;
			_isAttribute = false;
			return;
		}

		// if the user is ranging-out text
		if (lastRanged != null) {
			if (e.getButton() == MouseEvent.BUTTON1) {
				return;
			}

			Text ranged = lastRanged.copy();

			// if the user is cutting text from the item
			if (e.getButton() == MouseEvent.BUTTON2) {
				// Check if the user is trying to range an item for which they
				// do not have permission to do so... or it is the frame name
				if (lastRanged.Permission < Item.PERMISSION_FULL
						|| lastRanged.isFrameName()) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to cut range");
					lastRanged.clearSelection();
					FrameGraphics.Repaint();
					return;
				}
				// if the entire text is selected then pickup the item
				if (lastRanged.getSelectedSize() == lastRanged.getLength()) {
					pickup(lastRanged);
					lastRanged.clearSelection();
					FrameGraphics.Repaint();
					return;
				} else {
					ranged.setText(lastRanged.cutSelectedText());
				}
				// if the user is copying text from the item
			} else if (e.getButton() == MouseEvent.BUTTON3) {
				// Check if the user is trying to range an item for which they
				// do not have permission to do so... or it is the frame name
				if (lastRanged.Permission < Item.PERMISSION_COPY) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to copy range");
					lastRanged.clearSelection();
					FrameGraphics.Repaint();
					return;
				}
				ranged.setText(lastRanged.copySelectedText());
			}

			ranged.setLink(null);
			ranged.setParent(null);
			ranged.setPosition(DisplayIO.getMouseX(), DisplayIO.getMouseY());
			pickup(ranged);

			lastRanged.clearSelection();
			FrameGraphics.Repaint();
			return;
		}

		// if the user is cropping an image
		if (clickedOn != null && clickedOn == _lastCropped) {
			if (_lastCropped.isCropTooSmall()) {
				_lastCropped = null;
				// FrameGraphics
				// .WarningMessage("Crop cancelled because it was below the
				// minimum size");
			} else {
				Picture cropped = _lastCropped.copy();
				cropped.setParent(null);
				pickup(cropped);
			}
		}

		// if the user has cropped an image, either the above happend or this is
		// a no-op
		if (_lastCropped != null && !_lastCropped.isCropTooSmall()) {
			_lastCropped.clearCropping();
			_lastCropped = null;
			FrameGraphics.Repaint();
			return;
		}

		// if the user is left-clicking
		if (e.getButton() == MouseEvent.BUTTON1) {
			SessionStats.AddFrameEvent("Lu");
			leftButton(clickedOn);
			return;
		}

		if (e.getButton() == MouseEvent.BUTTON2) {
			SessionStats.AddFrameEvent("Mu");
			middleButton(clickedOn, clickedIn);
			return;
		}

		if (e.getButton() == MouseEvent.BUTTON3) {
			SessionStats.AddFrameEvent("Ru");
			rightButton(clickedOn, clickedIn);
			return;
		}

		// error, we should have returned by now
		System.out.println("Error: mouseReleased should have returned by now. "
				+ e);
	}

	/**
	 * This method handles all left-click actions
	 */
	public static void leftButton(Item clicked) {
		// if the user is pointing at something then either follow the link or
		// do TDFC
		if (clicked == null) {
			// Check if the user is nearby another item...
			int mouseX = DisplayIO.getMouseX();
			int mouseY = DisplayIO.getMouseY();
			// System.out.println(mouseX + "," + mouseY);
			for (Item i : DisplayIO.getCurrentFrame().getItems()) {
				if (i instanceof Text) {
					if (i.isNear(mouseX, mouseY)) {
						clicked = i;
						break;
					}
				}
			}
		}

		if (clicked instanceof Text) {
			if (((Text) clicked).getText().length() == 0)
				clicked = null;
		}

		if (clicked != null && !(clicked instanceof Line)
				&& !(clicked instanceof Dot)) {
			// check item permissions
			if (clicked.Permission < Item.PERMISSION_FOLLOW_LINKS) {
				FrameGraphics
						.DisplayMessage("Insufficient permissions to perform action on item");
				return;
			}

			Item clickedOn = clicked;

			// actions take priority
			if (_lastMouseClick != null && !_lastMouseClick.isShiftDown()
					&& clickedOn.getAction() != null) {
				clickedOn.performActions();
				return;
			} else if (clickedOn.getLink() != null) {
				// Dont save the frame if we are moving to an old version of
				// this frame
				// because everytime we save with the old tag... the frame is
				// backed up
				if (!clickedOn.isOldTag())
					FrameIO.SaveFrame(DisplayIO.getCurrentFrame());

				NavigationActions.setLastNavigationItem(clickedOn);
				load(clickedOn.getAbsoluteLink());
				// DisplayIO.UpdateTitle();
				return;
				// no link is found, perform TDFC
			} else {
				// check for TDFC permission
				if (clicked.Permission < Item.PERMISSION_TDFC) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to TDFC from that item");
					return;
				}

				if (clickedOn.isOldTag())
					return;

				// if the user is clicking on the frame name
				if (clickedOn.isFrameName()) {
					Frame next = FrameIO.LoadNext();
					FrameUtils.DisplayFrame(next, true);
					return;
				}
				try {
					tdfc(clickedOn);
				} catch (RuntimeException e) {
					FrameGraphics.ErrorMessage(e.getMessage());
				}
				return;
			}
			// if user is not pointing at something,this is a back
		} else {
			back();
			return;
		}
	}

	private static boolean doMerging(Item clicked) {
		if (clicked == null)
			return false;

		// Brook: widgets do not merge
		if (clicked instanceof WidgetCorner)
			return false;

		// Brook: widgets do not merge
		if (clicked instanceof WidgetEdge)
			return false;

		// System.out.println(Frame.FreeItems.size());
		if (isRubberBandingCorner()) {
			if (clicked.isLineEnd()
					|| clicked.getAllConnected().contains(
							Frame.getItemAttachedToCursor())) {
				return true;
			}
		}

		if (Frame.FreeItems.size() > 2)
			return false;

		Item attachedToCursor = Frame.getItemAttachedToCursor();

		if (clicked instanceof Text
				&& !(attachedToCursor instanceof Text || attachedToCursor
						.isLineEnd())) {
			return false;
		}

		return true;
	}

	/**
	 * This method handles all middle-click actions
	 */
	private static void middleButton(Item clicked, List<Item> clickedIn) {
		// if the cursor has Items attached
		if (Frame.itemAttachedToCursor()) {
			// if the user is pointing at something, merge the items (if
			// possible)
			if (doMerging(clicked)) {
				// check permissions
				if (clicked.Permission < Item.PERMISSION_FULL
						&& clicked.getParent() != null
						&& clicked.getParent().getNameItem() != clicked) {
					FrameGraphics.DisplayMessage("Insufficient permission");
					return;
				}

				Item merger = Frame.FreeItems.get(0);
				assert (merger != null);
				Collection<Item> left = merge(Frame.FreeItems, clicked);
				anchor(left);
				Frame.FreeItems.clear();
				FrameGraphics.Repaint();
				updateCursor();
				return;
				// otherwise, anchor the items
			} else {
				// if a line is being rubber-banded, check for auto
				// straightening
				anchor(Frame.FreeItems);
				Frame.FreeItems.clear();
				updateCursor();
				_offX = _offY = 0;
				return;
			}
			// otherwise if the user is pointing at something,pick it up
		} else if (clicked != null) {
			// check permissions
			if (clicked.Permission < Item.PERMISSION_FULL) {
				FrameGraphics
						.DisplayMessage("Insufficient permission to pick up item");
				return;
			}

			// BROOK: WIDGET RECTANGLES DONT ALLOW DISCONNECTION
			if (clicked instanceof Line && !(clicked instanceof WidgetEdge)) {
				// Check if within 20% of the end of the line
				Line l = (Line) clicked;
				Item toDisconnect = l.getEndPointToDisconnect(_lastMouseClick
						.getX(), _lastMouseClick.getY());

				if (toDisconnect == null) {
					pickup(clicked);
				} else {
					if (toDisconnect.getSelectedMode() == Item.SelectedMode.Normal) {
						DisplayIO.setCursorPosition(toDisconnect.getPosition(),
								false);
						pickup(toDisconnect);
					} else {
						List<Line> lines = toDisconnect.getLines();
						// This is to remove constraints from single lines
						// with constraints...
						// ie. partially deleted rectangles
						if (lines.size() == 1) {
							toDisconnect.removeAllConstraints();

							DisplayIO.setCursorPosition(toDisconnect
									.getPosition(), false);
							pickup(toDisconnect);
						} else {
							// If we are then detatch the line and pick up its
							// end point...
							Frame currentFrame = DisplayIO.getCurrentFrame();
							Item newPoint = null;

							// If the point we are disconnecting is text...
							// Then we want to leave the text behind
							// And disconnect a point
							if (toDisconnect instanceof Text) {
								newPoint = new Dot(toDisconnect.getX(),
										toDisconnect.getY(), -1);
								Item.DuplicateItem(toDisconnect, newPoint);
							} else {
								newPoint = toDisconnect.copy();
							}

							currentFrame.addItem(newPoint);
							// remove the current item from the connected
							// list for this item
							l.replaceEnd(toDisconnect, newPoint);
							// remove unneeded constrains
							newPoint.removeAllConstraints();

							toDisconnect
									.setSelectedMode(Item.SelectedMode.None);
							DisplayIO.setCursorPosition(toDisconnect
									.getPosition(), false);
							pickup(newPoint);
							ItemUtils.EnclosedCheck(toDisconnect
									.getParentOrCurrentFrame().getItems());
						}
					}
				}
			} else {
				if (clicked.isLineEnd()) {
					DisplayIO.setCursorPosition(clicked.getPosition(), false);
				}
				pickup(clicked);
			}
			// otherwise create a line
		} else if (clickedIn != null) {
			ArrayList<Item> toPickup = new ArrayList<Item>(clickedIn.size());
			for (Item ip : clickedIn)
				if (ip.Permission >= Item.PERMISSION_FULL)
					toPickup.add(ip);
			pickup(toPickup);
			// otherwise the user is creating a line
		} else {
			Item on = FrameUtils.onItem(DisplayIO.getCurrentFrame(), MouseX,
					MouseY);
			Line newLine = createLine();
			SessionStats.CreatedItems(newLine.getAllConnected());

			// if the user can spot-weld, show the virtual spot
			if (on instanceof Line) {
				Line line = (Line) on;
				newLine.setThickness(line.getThickness());
				newLine.setLinePattern(line.getLinePattern());
				newLine.setColor(line.getColor());
				line.forceMerge(newLine.getStartItem(), MouseX, MouseY);
			}
			return;
		}
		SessionStats.MovedItems(Frame.FreeItems);
	}

	private static Item getFirstFreeLineEnd() {
		for (Item i : Frame.FreeItems)
			if (i.isLineEnd())
				return i;
		return null;
	}

	private static boolean isRubberBandingCorner() {
		return getShapeCorner(Frame.FreeItems) != null;
	}

	/**
	 * Gets the rectangle corner from the list of items that are part of a
	 * rectangle.
	 * 
	 * @param partialRectangle
	 *            a corner and its two connecting lines.
	 * @return the rectangle corner or null if the list of items is not part of
	 *         a rectangle.
	 */
	private static Item getShapeCorner(List<Item> partialRectangle) {
		if (partialRectangle.size() < 3)
			return null;
		Item lineEnd = null;
		// only one lineEnd will be present for rectangles
		// All other items must be lines
		for (Item i : partialRectangle) {
			if (i.isLineEnd()) {
				if (lineEnd == null) {
					lineEnd = i;
				} else {
					return null;
				}
			} else if (!(i instanceof Line)) {
				return null;
			}
		}
		// if this is at least the corner of two connected lines
		if (lineEnd != null && lineEnd.getAllConnected().size() >= 5)
			return lineEnd;

		return null;
	}

	/**
	 * This method handles all right-click action
	 */
	private static void rightButton(Item clicked, List<Item> clickedIn) {
		// if the cursor has Items attached, then anchor a copy of them
		List<Item> copies = null;
		if (Frame.itemAttachedToCursor()) {
			// if the user is clicking on something, merge the items
			// unless it is a point onto somethin other than a lineEnd or a dot
			if (clicked != null
					&& (!(Frame.getItemAttachedToCursor() instanceof Dot)
							|| clicked instanceof Dot || clicked.isLineEnd())) {
				// check permissions
				if (clicked.Permission < Item.PERMISSION_FULL
						&& clicked.getParent().getNameItem() != clicked) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to merge items");
					return;
				}
				if (clicked instanceof Text || clicked instanceof Dot) {
					if (isRubberBandingCorner()) {
						// Move the cursor so that the copy is exactly the
						// same as the shape that was anchored
						DisplayIO.setCursorPosition(clicked.getPosition());
						Item d = getFirstFreeLineEnd();
						// get a copy of all enclosed items before merging
						// lineEnds
						Collection<Item> items = FrameUtils.getItemsEnclosedBy(
								DisplayIO.getCurrentFrame(), d
										.getEnclosedShape());
						// If its not an enclosed shape then pick up the
						// connected shape
						if (items == null || items.size() == 0) {
							items = d.getAllConnected();
						} else {
							// For some reason the item that was clicked ends up
							// in the enclosure and needs to be removed
							items.removeAll(clicked.getConnected());
							// the item that was the origin of the enclosed
							// shape used to create the enclosure does not get
							// returned from getItemsEnclosedBy to the enclosure
							// so it must be added
							items.addAll(d.getConnected());
						}

						Collection<Item> toCopy = new LinkedHashSet<Item>();

						for (Item ip : items) {
							if (ip.Permission >= Item.PERMISSION_COPY)
								toCopy.add(ip);
						}
						copies = copy(toCopy);
						// Now do the merging
						Collection<Item> remain = merge(Frame.FreeItems,
								clicked);
						// anchor the points
						anchor(remain);
						Frame.FreeItems.clear();
						pickup(copies);
						// line onto something
					} else if (Frame.FreeItems.size() == 2) {
						copies = ItemUtils.UnreelLine(Frame.FreeItems);
						Collection<Item> leftOver = merge(Frame.FreeItems,
								clicked);
						anchor(leftOver);
						if (copies == null)
							copies = copy(Frame.FreeItems);
						Frame.FreeItems.clear();
						for (Item i : copies)
							i.setOffset(0, 0);
						// need to move to prevent cursor dislocation
						move(copies);
						pickup(copies);
						// point onto point
					} else if (Frame.FreeItems.size() == 1) {
						copies = copy(Frame.FreeItems);
						Collection<Item> remain = merge(copies, clicked);

						// ignore items that could not be merged.
						anchor(remain);
					} else {
						stampItemsOnCursor();
						copies = Frame.FreeItems;
					}
				} else {
					copies = ItemUtils.UnreelLine(Frame.FreeItems);
					if (copies == null)
						copies = copy(Frame.FreeItems);
					for (Item i : copies) {
						i.setOffset(0, 0);
					}
					anchor(Frame.FreeItems);
					Frame.FreeItems.clear();
					pickup(copies);
				}
				// otherwise, anchor the items
			} else {
				// check if this is anchoring a rectangle
				if (isRubberBandingCorner()) {
					Item d = getFirstFreeLineEnd();
					// anchor the points
					anchor(Frame.FreeItems);
					Frame.FreeItems.clear();
					updateCursor();
					// pick up a copy of all enclosed items
					List<Item> enclosedItems = FrameUtils.getItemsEnclosedBy(
							DisplayIO.getCurrentFrame(), d.getEnclosedShape());
					if (enclosedItems != null) {
						enclosedItems.removeAll(d.getAllConnected());
						List<Item> toCopy = getFullyEnclosedItems(enclosedItems);

						if (toCopy.size() > 0) {
							// Find the closest item to the mouse cursor
							double currentX = DisplayIO.getMouseX();
							double currentY = DisplayIO.getMouseY();
							Item closest = null;
							double shortestDistance = Double.MAX_VALUE;
							for (int i = 0; i < toCopy.size(); i++) {
								Item next = toCopy.get(i);
								if (next instanceof Line)
									continue;
								double distance = Point.distance(currentX,
										currentY, next.getX(), next.getY());
								if (distance < shortestDistance) {
									shortestDistance = distance;
									closest = next;
								}
							}
							// Move the cursor to closest item
							DisplayIO.setCursorPosition(closest.getPosition());
							// Pickup copy of the stuff inside the rectange
							copies = copy(toCopy);
							pickup(copies);
							// Remove the rectangle
							d.getParentOrCurrentFrame().removeAllItems(
									d.getAllConnected());
						} else {
							// Pick up a copy of the rectangle
							copies = copy(d.getAllConnected());
							pickup(copies);
						}
					}
				} else {
					if (Frame.FreeItems.size() == 2) {
						if (clicked != null) {
							Collection<Item> leftOver = merge(Frame.FreeItems,
									clicked);
							anchor(leftOver);
						}
						// This is executed when the user is putting down a line
						// endpoint and unreeling. ie. Normal unreeling
						copies = ItemUtils.UnreelLine(Frame.FreeItems);

						if (copies == null)
							copies = copy(Frame.FreeItems);
						anchor(Frame.FreeItems);
						for (Item i : copies)
							i.setOffset(0, 0);
						// need to move to prevent cursor dislocation
						move(copies);
						pickup(copies);
					} else if (_extrude) {
						List<Item> originals = new ArrayList<Item>();
						// remove any lines that dont have both endpoints
						// floating
						for (Item i : Frame.FreeItems) {
							if (i.isFloating())
								originals.add(i);
						}
						if (copies == null)
							copies = ItemUtils.CopyItems(originals, _extrude);
						for (Item i : copies)
							i.setOffset(0, 0);
						anchor(Frame.FreeItems);
						// Move isnt working right for extruding!!
						// move(copies);
						pickup(copies);
					} else {
						stampItemsOnCursor();
						copies = Frame.FreeItems;
					}
				}
			}
		} else {
			// if the user is pointing at something, this is a copy
			if (clicked != null) {
				// check permissions
				if (clicked.Permission < Item.PERMISSION_COPY) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to copy item");
					return;
				}

				copies = ItemUtils.UnreelLine(clicked);
				// Copies will NOT be null if the user right clicked on a point
				if (copies == null) {
					Collection<Item> originals = clicked.getConnected();
					copies = ItemUtils.CopyItems(originals, _extrude);
					// if this is the title of the frame, link it to the frame
					if (originals.size() == 1 && copies.size() == 1) {
						Item copy = copies.get(0);
						Item original = originals.iterator().next();
						if (original.getLink() == null && original.isFrameTitle()) {
							// save the frame after copying
							// i.getParent().setChanged(true);
							copy.setLink(original.getParentOrCurrentFrame().getName());
						}
					}

					if (!_extrude)
						clearParent(copies);
				}

				pickup(copies);
			} else {
				// if user is pointing in a closed shape, copy the items inside
				if (clickedIn != null) {
					// Set the selection mode for the items that were clicked in
					List<Item> enclosed = getFullyEnclosedItems(clickedIn);
					if (enclosed.size() == 0) {
						FrameGraphics
								.DisplayMessage("Insufficient permission to copy items");
					} else {
						copies = copy(enclosed);
						// TODO figure out how to get copies to be SELECTED when
						// they are anchored
						// for (Item i : copies) {
						// if (i.isEnclosed()) {
						// i.setSelectedMode(Item.SelectedMode.Enclosed);
						// i.setSelectedMode(Item.SelectedMode.None);
						// }
						// }
						clearParent(copies);
						pickup(copies);
					}
					// otherwise, create a rectangle
				} else {
					copies = new ArrayList<Item>();
					Item[] d = new Item[RECTANGLE_CORNERS];
					// create dots
					Frame current = DisplayIO.getCurrentFrame();
					for (int i = 0; i < d.length; i++) {
						d[i] = current.createDot();
						copies.add(d[i]);
					}

					// create lines
					copies.add(new Line(d[0], d[1], current.getNextItemID()));
					copies.add(new Line(d[1], d[2], current.getNextItemID()));
					copies.add(new Line(d[2], d[3], current.getNextItemID()));
					copies.add(new Line(d[3], d[0], current.getNextItemID()));

					new Constraint(d[0], d[1], current.getNextItemID(),
							Constraint.HORIZONTAL);
					new Constraint(d[2], d[3], current.getNextItemID(),
							Constraint.HORIZONTAL);
					new Constraint(d[1], d[2], current.getNextItemID(),
							Constraint.VERTICAL);
					new Constraint(d[3], d[0], current.getNextItemID(),
							Constraint.VERTICAL);

					anchor(new ArrayList<Item>(copies));
					pickup(d[3]);
					SessionStats.CreatedItems(copies);
					copies.clear();
				}
			}
		}
		SessionStats.CopiedItems(copies);
		updateCursor();
		FrameGraphics.Repaint();
	}

	/**
	 * 
	 */
	private static void stampItemsOnCursor() {
		List<Item> copies = copy(Frame.FreeItems);
		// MIKE: what does the below 2 lines do?
		for (Item i : copies)
			i.setOffset(0, 0);
		// The below code has a little problem withflicker when stamp
		// and dragging
		move(Frame.FreeItems);
		anchor(copies);
	}

	/**
	 * @param enclosedItems
	 * @return
	 */
	private static List<Item> getFullyEnclosedItems(List<Item> enclosure) {
		// copy the enclosedItems because the list will be modified
		List<Item> enclosedItems = new ArrayList<Item>(enclosure);
		List<Item> toCopy = new ArrayList<Item>(enclosedItems.size());
		while (enclosedItems.size() > 0) {
			Item i = enclosedItems.get(0);
			if (i.Permission >= Item.PERMISSION_COPY) {
				Collection<Item> items = i.getAllConnected();
				// Only copy if the entire shape is enclosed
				if (enclosedItems.containsAll(items)) {
					toCopy.addAll(items);
				}
				enclosedItems.removeAll(items);
			} else {
				enclosedItems.remove(0);
			}
		}
		return toCopy;
	}

	/**
	 * Marks the items as not belonging to any specific frame. When picking up
	 * items the parent will be automatically cleared for items on the current
	 * frame but not for overlay items. This method ensures that overlay items
	 * will also be cleared. This is useful when picking up copies of items from
	 * an overlay (with the right mouse button) to ensure that the copy will be
	 * anchored on the current frame rather than the overlay. When items are
	 * picked up with the middle button clearParent should NOT be called.
	 * 
	 * @param items
	 *            to have their parent cleared
	 */
	private static void clearParent(List<Item> items) {
		for (Item i : items)
			i.setParent(null);
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void mouseDragged(MouseEvent e) {
		// System.out.println("MouseDragged");

		// check if user is dragging across a text item
		if (_lastRanged != null) {
			// System.out.println(MouseY - e.getY());

			MouseX = e.getX();
			MouseY = e.getY();

			int distance = _lastRanged.getY() - DisplayIO.getMouseY();
			if (distance <= 0)
				distance = DisplayIO.getMouseY() - _lastRanged.getY()
						- _lastRanged.getBoundsHeight();

			if (distance > UserSettings.NoOpThreshold) {
				_lastRanged.clearSelectionEnd();
				_isNoOp = true;
			} else {
				// update the ranged section
				_lastRanged.setSelectionEnd(DisplayIO.getMouseX(), DisplayIO
						.getMouseY());
				_isNoOp = false;
			}

			DisplayIO.setTextCursor(_lastRanged.getSize());
			FrameGraphics.Repaint();
			return;
		}

		// if the user is dragging across a picture
		if (_lastCropped != null) {
			MouseX = e.getX();
			MouseY = e.getY();

			// update the ranged section
			_lastCropped.setEndCrop(DisplayIO.getMouseX(), DisplayIO
					.getMouseY());
			DisplayIO.setCursor(Item.CROP_CURSOR);
			FrameGraphics.Repaint();
			return;
		}

		if (_mouseDown == MouseEvent.BUTTON2 && _lastClickedOn == null
				&& _lastClickedIn == null) {
			Item on = FrameUtils.onItem(DisplayIO.getCurrentFrame(), e.getX(),
					e.getY());
			// if the user can spot-weld, show the virtual spot
			if (Frame.FreeItems.size() == 0 && on instanceof Line) {
				Line line = (Line) on;
				line.showVirtualSpot(e.getX(), e.getY());
			}
		}

		/*
		 * // Euclidean distance int distance = (int)
		 * Math.sqrt(Math.pow(Math.abs(MouseX - e.getX()), 2) +
		 * Math.pow(Math.abs(MouseY - e.getY()), 2));
		 * 
		 * if (distance > UserSettings.NoOpThreshold)
		 */
		// Use the below calculation for better speed. If it causes problems
		// switch back to the Euclidean distance calculation
		if (Math.abs(MouseX - e.getX()) > UserSettings.NoOpThreshold
				|| Math.abs(MouseY - e.getY()) > UserSettings.NoOpThreshold)
			_isNoOp = true;

		FrameGraphics.Repaint();
	}

	MouseEvent _lastMouseEvent = null;

	private static Integer LastRobotX = null;

	private static Integer LastRobotY = null;

	// For some reason... sometimes the mouse move gets lost when moving the
	// mouse really quickly after clicking...
	// Use this timer to make sure it gets reset eventually if the Robot
	// generated event never arrives.
	private static Timer _RobotTimer = new Timer(200, new ActionListener() {
		public void actionPerformed(ActionEvent ae) {
			_RobotTimer.stop();
			LastRobotX = null;
			LastRobotY = null;
			// System.out.println("RobotTimer");
		}
	});

	public static void setLastRobotMove(int x, int y) {
		// Make sure the system is in the right state while waiting for the
		// Robots event to arrive.
		MouseX = x;
		MouseY = y;
		// System.out.println("MouseMoved: " + MouseX + "," + MouseY + " " +
		// System.currentTimeMillis());
		LastRobotX = x;
		LastRobotY = y;
		_RobotTimer.start();
	}

	public static boolean isWaitingForRobot() {
		return LastRobotX != null;
	}

	/**
	 * Updates the stored mouse position and highlights any items as necessary.
	 */
	public void mouseMoved(MouseEvent e) {
		// System.out.println(e.getX() + "," + e.getY() + " " + e.getWhen());
		if (LastRobotX != null) {
			// Wait until the last Robot mouse move event arrives before
			// processing other events
			if (/* Frame.FreeItems.size() == 0 || */
			(LastRobotX == e.getX() && LastRobotY == e.getY())) {
				LastRobotX = null;
				LastRobotY = null;
				_RobotTimer.stop();
			} else {
				// System.out.println("Ignored: " + Frame.FreeItems.size());
				return;
			}
		}

		MouseX = e.getX();
		MouseY = e.getY();

		if (_lastMouseEvent == null)
			_lastMouseEvent = e;

		/*
		 * int distance = (int) Math.sqrt(Math.pow(Math
		 * .abs(_lastMouseEvent.getX() - e.getX()), 2) +
		 * Math.pow(Math.abs(_lastMouseEvent.getY() - e.getY()), 2));
		 */

		// Just do the sum of the distance moved in x and y direction for
		// speed!!
		/*
		 * int distance = Math.abs(_lastMouseEvent.getX() - e.getX()) +
		 * Math.abs(_lastMouseEvent.getY() - e.getY());
		 */

		boolean checkHighlight = true;

		// Mike: This code checks if the mouse is moving above a
		// threshold speed. If so it doesnt execute the code that checks if
		// items underneath the cursor need to be highlighted.
		/*
		 * if (Frame.itemAttachedToCursor()) { long time = e.getWhen() -
		 * _lastMouseEvent.getWhen(); // System.out.println(time + " D:" +
		 * distance); if (time > 0 && distance * 3 > time) { checkHighlight =
		 * false; // System.out.println("Dont check highlight!"); if
		 * (_lastHighlightedItem != null && !_lastHoldsHighlight) {
		 * if(_lastHighlightedItem instanceof Text) {
		 * FrameGraphics.ChangeSelectionMode(_lastHighlightedItem,
		 * Item.SelectedMode.None); } } } }
		 */

		_lastMouseEvent = e;

		if (checkHighlight) {
			// ByMike: Get the item the mouse is hovering over
			Item click = FrameUtils.getCurrentItem();
			Item on = null;
			// System.out.println(click);
			if (click != null) {
				on = click;
				// set the context
				if (on instanceof Line)
					_context = CONTEXT_AT_LINE;
				else if (on instanceof Dot)
					_context = CONTEXT_AT_DOT;
				else if (on instanceof Text) {
					// Checks that we are actually pointing on a character
					// not just space in the text box's bounding box
					if (((Item) on).contains(DisplayIO.getMouseX(), DisplayIO
							.getMouseY())) {
						_context = CONTEXT_AT_TEXT;
					} else {
						// if we are pointing in space in the text box... it is
						// the same as pointing in freespace
						on = null;
						_context = CONTEXT_FREESPACE;
					}
				}

				_alpha = 60;
			} else {
				_context = CONTEXT_FREESPACE;
				_alpha = -1;
			}

			// if the user is pointing at an item, highlight it
			if (on != null && !Frame.FreeItems.contains(on)) {
				// if the user can spot-weld, show the virtual spot
				if (Frame.FreeItems.size() == 2 && on instanceof Line) {
					Line line = (Line) on;
					Item freeItem0 = Frame.FreeItems.get(0);
					Item freeItem1 = Frame.FreeItems.get(1);
					Item lineEnd = freeItem0.isLineEnd() ? freeItem0
							: (freeItem1.isLineEnd() ? freeItem1 : null);
					if (lineEnd != null) {
						line.showVirtualSpot(lineEnd, DisplayIO.getMouseX(),
								DisplayIO.getMouseY());
					} else
						// The user is pointing at another point or text item
						// etc
						FrameGraphics.ChangeSelectionMode(on,
								Item.SelectedMode.Normal);
				} else {
					// FrameGraphics.ChangeSelectionMode(on,
					// Item.SelectedMode.Connected);
					// TODO: The method below is for the most part redundant
					on = FrameGraphics.Highlight(on);
				}
				// if the last item highlighted is still highlighted, clear it
				if (_lastHoldsHighlight) {
					_lastHoldsHighlight = false;
					for (Item i : DisplayIO.getCurrentFrame().getItems())
						if (i.isHighlighted() && i != on)
							FrameGraphics.ChangeSelectionMode(i,
									Item.SelectedMode.None);
				}

				// if the user is not pointing at an item, check for enclosure
				// highlighting
			} else if (on == null) {
				Collection<Item> enclosure = FrameUtils.getEnclosingDots();
				if (enclosure != null && enclosure.size() > 1) {
					Item firstLineEnd = enclosure.iterator().next();
					if (firstLineEnd.getLines().size() > 1 &&
					// check that the enclosure is not part of a point being
							// dragged in space
							!ContainsOneOf(enclosure, Frame.FreeItems)) {
						on = firstLineEnd.getLines().get(0);
						// System.out.println(on == null ? "Null" :
						// on.toString());
						FrameGraphics.ChangeSelectionMode(on,
								Item.SelectedMode.Enclosed);
					}
				} else if (_lastHighlightedItem != null) {
					// System.out.println("LastHighlightedItem");
					_lastHoldsHighlight = false;
				}
			}

			// disable cursor changes when the cursor has items attached
			if (Frame.itemAttachedToCursor()
					&& DisplayIO.getCursor() != Item.TEXT_CURSOR)
				_forceArrowCursor = false;

			if (_lastHighlightedItem != null && _lastHighlightedItem != on
					&& !_lastHoldsHighlight) {
				// Only want to turn off the highlighting only if
				// the last highlighted item is not connected to the currentItem
				// Otherwise we get flickering in transition from connected to
				// normal mode while moving the cursor along a line.
				if (on == null
						|| !_lastHighlightedItem.getAllConnected().contains(on))
					FrameGraphics.ChangeSelectionMode(_lastHighlightedItem,
							Item.SelectedMode.None);
			}

			_lastHighlightedItem = on;

		}

		if (Frame.itemAttachedToCursor()) {
			move(Frame.FreeItems);
			// System.out.println(Frame.FreeItems.size());
		}

		if (_forceArrowCursor)
			updateCursor();

		// if (_lastItem == null)
		// DisplayIO.UpdateTitle();

		_forceArrowCursor = true;
	}

	private boolean ContainsOneOf(Collection<Item> enclosure,
			Collection<Item> freeItems) {
		if (freeItems == null)
			return false;
		for (Item i : freeItems) {
			if (enclosure.contains(i))
				return true;
		}
		return false;
	}

	/**
	 * Checks if lines are being rubber banded.
	 * 
	 * @return true if the user is rubberBanding one or more lines
	 */
	private static boolean rubberBanding() {
		if (Frame.FreeItems.size() != 2) {
			return false;
		}

		// if rubber-banding, there will be 1 lineend and the rest will be lines
		boolean foundLineEnd = false;
		for (Item i : Frame.FreeItems) {
			if (i.isLineEnd()) {
				if (foundLineEnd) {
					return false;
				}
				foundLineEnd = true;
			} else if (!(i instanceof Line)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Updates the current mouse cursor to whatever it should be. i.e. Hidden
	 * when rubber-banding lines, otherwise default (arrow)
	 */
	public static void updateCursor() {
		Item on = FrameUtils.getCurrentItem();
		if (rubberBanding()) {
			DisplayIO.setCursor(Item.HIDDEN_CURSOR);
			return;
		}/*
			 * else if(on != null && on instanceof Text){
			 * DisplayIO.setTextCursor(((Item) on).getSize()); }
			 */else {
			DisplayIO.setCursor(Item.DEFAULT_CURSOR);
		}
	}

	public static void setHighlightHold(boolean hold) {
		_lastHoldsHighlight = hold;
	}

	public static void resetOffset() {
		if (Frame.itemAttachedToCursor()) {
			_offX = DisplayIO.getMouseX() - Frame.FreeItems.get(0).getX()
					+ Frame.FreeItems.get(0).getOffset().x;
			_offY = MouseY - Frame.FreeItems.get(0).getY()
					+ Frame.FreeItems.get(0).getOffset().y;
		}
	}

	private static void move(Collection<Item> toMove) {

		// Gets the origin of the first item to move
		int xPos = (DisplayIO.getMouseX() - _offX);

		Item firstDot = toMove.iterator().next();

		int deltax = firstDot.getX() - xPos;
		int deltay = firstDot.getY() - (MouseY - _offY);

		for (Item move : toMove) {
			move.setPosition(move.getX() - deltax, move.getY() - deltay);

			if (move instanceof Text) {
				((Text) move).setAlpha(_alpha);
			}
		}

		FrameGraphics.Repaint();
	}

	private static void load(String toLoad) {
		if (FrameIO.isValidFrameName(toLoad)) {
			FrameUtils.DisplayFrame(toLoad);
		} else {
			FrameGraphics.ErrorMessage(toLoad + " is not a valid frame name.");
		}
	}

	private static void back() {
		DisplayIO.Back();

		// repaint things if necessary
		if (Frame.itemAttachedToCursor())
			move(Frame.FreeItems);
	}

	/**
	 * Returns true if the mouse moved during TDFC. This will happen if there is
	 * a start annotation item on the frame.
	 * 
	 * @param linker
	 * @return
	 */
	public static boolean tdfc(Item linker) throws RuntimeException {
		if (linker instanceof Line)
			return false;

		// if this is a non-usable item
		if (linker.getID() < 0)
			return false;

		boolean mouseMoved;

		linker.getParent().setChanged(true);

		Frame next = FrameIO.CreateNewFrame(linker);

		linker.setLink("" + next.getNumber());

		for (Item i : next.getBodyTextItems(true)) {
			// Set the link for @Parent annotation item if one
			if (ItemUtils.startsWithTag(i, ItemUtils.TAG_PARENT)
					&& i.getLink() == null) {
				Frame parent = linker.getParentOrCurrentFrame();
				i.setLink(parent.getName());
			} else if (ItemUtils.startsWithTag(i, ItemUtils.TAG_BACKUP, false)) {
				// Delink backup tag if it is on the frame
				i.setLink(null);
			}
		}

		FrameUtils.DisplayFrame(next, true);
		FrameUtils.setTdfcItem(linker);

		mouseMoved = next.moveMouseToDefaultLocation();
		// this needs to be done if the user doesnt move the mouse before doing
		// tdfc while the cursor is set to the text cursor
		DisplayIO.setCursor(Item.DEFAULT_CURSOR);
		// This needs to be done in case there was a @start on the frame which
		// triggers changed to be set to true when it should stay as false
		next.setChanged(false);
		return mouseMoved;
	}

	/**
	 * Creates a new Text item and fills it with particular attributes extracted
	 * from the given Item. Note: Users always have permission to extract
	 * attributes, so it is not checked.
	 * 
	 * @param toExtract
	 *            Item containing the Item to extract the attributes from.
	 */
	private static void extractAttributes(Item toExtract) {
		if (toExtract == null || toExtract == null)
			return;

		if (Frame.itemAttachedToCursor())
			return;

		Item attribs;
		Item item = toExtract;
		// Extract the frames attributes when the user clicks on the frame name
		if (item.isFrameName())
			attribs = AttributeUtils.ExtractAttributes(item.getParent());
		else
			attribs = AttributeUtils.ExtractAttributes(item);

		if (attribs == null)
			FrameGraphics
					.DisplayMessage("All attributes of that item are default values.");
		else {
			// Give the attribute text item the color of the item for which
			// attributes are being extracted.
			// attribs.setColor(item.getColor());
			pickup(attribs);
		}
	}

	public static void delete(Item toDelete) {
		FrameUtils.LastEdited = null;
		_offX = _offY = 0;

		// check if the user is pointing at the frame's framename
		if (toDelete != null
				&& toDelete == DisplayIO.getCurrentFrame().getNameItem()) {
			DisplayIO.getCurrentFrame().clear();
			FrameGraphics.Repaint();
			return;
		}

		// if the user is deleting items attached to the cursor
		if (Frame.itemAttachedToCursor()) {
			// if this is an item-swap
			if (Frame.FreeItems.size() == 1
					&& Frame.FreeItems.get(0) instanceof Text
					&& toDelete != null && toDelete instanceof Text) {

				// check permissions
				if (toDelete.Permission < Item.PERMISSION_FULL) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to swap Item text");
					return;
				}
				Text anchored = (Text) toDelete;
				Text free = (Text) Frame.FreeItems.get(0);
				SessionStats.DeletedItem(free);
				// List<String> temp = anchored.getText();
				anchored.setTextList(free.getTextList());

				// free.setTextList(temp);
				Frame.FreeItems.clear();

				anchored.getParent().setChanged(true);

				// update the offset since the text has changed
				_offX = DisplayIO.getMouseX() - anchored.getX()
						+ anchored.getOffset().x;
				_offY = MouseY - anchored.getY() + anchored.getOffset().y;
			} else {
				deleteItems(Frame.FreeItems);
			}
			// reset the mouse cursor
			updateCursor();
			FrameGraphics.Repaint();
			return;
			// the user is not pointing at an item
		} else if (toDelete == null) {

			// if the user is pointing inside a closed shape, delete all
			// items inside it
			List<Item> items = FrameUtils.getCurrentItems(null);

			if (items != null) {
				Collection<Item> toRemove = new LinkedHashSet<Item>(items
						.size());
				for (Item ip : items) {
					if (ip.Permission >= Item.PERMISSION_FULL) {
						// Only include lines if one of their enpoints are also
						// being removed
						if (ip instanceof Line) {
							Line l = (Line) ip;
							if (!items.contains(l.getEndItem())
									&& !items.contains(l.getStartItem()))
								continue;
						}
						toRemove.add(ip);
					}
				}

				deleteItems(toRemove);

				// reset the mouse cursor
				updateCursor();
				FrameGraphics.Repaint();

				// otherwise this is an undo command
			} else {
				DisplayIO.getCurrentFrame().undo();
			}
			return;
			// this is a delete command
		} else {
			// check permissions
			if (toDelete.Permission < Item.PERMISSION_FULL) {
				FrameGraphics
						.DisplayMessage("Insufficient permission to delete item");
				return;
			}

			Frame current = toDelete.getParent();

			current.setChanged(true);

			Collection<Item> toUndo = null;
			if (toDelete.isLineEnd()) {
				toUndo = deleteLineEnd(toDelete);
			} else {
				toUndo = copy(toDelete.getConnected());
			}
			SessionStats.DeletedItems(toUndo);
			current.addAllToUndo(toUndo);
			current.removeAllItems(toDelete.getConnected());
			// reset the mouse cursor
			updateCursor();
			ItemUtils.EnclosedCheck(current.getItems());
			FrameGraphics.Repaint();
		}
		// check
		DisplayIO.setCursor(Item.DEFAULT_CURSOR);
	}

	private static void deleteItems(Collection<Item> itemList) {
		SessionStats.DeletedItems(itemList);
		List<Frame> modifiedFrames = new LinkedList<Frame>();
		// Get a list of all the modified frames
		for (Item i : itemList) {
			Frame parent = i.getParent();
			if (parent != null)
				modifiedFrames.add(parent);
		}
		// If they are all free items then add the current frame
		if (modifiedFrames.size() == 0) {
			modifiedFrames.add(DisplayIO.getCurrentFrame());
		}

		Collection<Item> toUndo = new LinkedHashSet<Item>();
		// disconnect any connected items
		for (Item i : itemList) {
			if (i.getLines().size() > 0) {
				Collection<Item> toDelete = deleteLineEnd(i);
				// add the copied items to the undo stack
				for (Item itemToUndo : toDelete)
					if (!toUndo.contains(itemToUndo))
						toUndo.add(itemToUndo);
			} else if (!toUndo.contains(i)) {
				toUndo.add(i); // Why was is this a copy
			}
		}
		for (Frame f : modifiedFrames) {
			f.removeAllItems(itemList);
			ItemUtils.EnclosedCheck(f.getItems());
		}
		// TODO: How should undelete deal with undo when items are removed from
		// the current frame as well as the overlay frame
		DisplayIO.getCurrentFrame().addAllToUndo(itemList);
		itemList.clear();
	}

	private static Collection<Item> deleteLineEnd(Item lineEnd) {

		if (lineEnd instanceof WidgetCorner) { // Brook

			WidgetCorner wc = (WidgetCorner) lineEnd;
			List<Item> wccopy;

			try {
				wccopy = wc.getWidgetSource().copy().getItems();
			} catch (InteractiveWidgetNotAvailableException e) {
				e.printStackTrace();
				return new LinkedList<Item>();
			} catch (InteractiveWidgetInitialisationFailedException e) {
				e.printStackTrace();
				return new LinkedList<Item>();
			}

			if (lineEnd.getParent() != null) {
				lineEnd.getParent().removeAllItems(
						wc.getWidgetSource().getItems());
			}
			return wccopy;
		} else {

			// // create a backup copy of the dot and its lines
			// List<Item> copy = copy(lineEnd.getConnected());
			//			
			// // Remove lines from their anchored dots
			// // note: the line is kept so that it can be properly restored
			// for (Item ic : copy) {
			// if (ic instanceof Line) {
			// Line line = (Line) ic;
			// // Remove the line from the item that is not the copy of the
			// // line end being deletedF
			// if (!copy.contains(line.getStartItem()))
			// line.getStartItem().removeLine(line);
			// if (!copy.contains(line.getEndItem()))
			// line.getEndItem().removeLine(line);
			// }
			// }

			Collection<Item> copy = lineEnd.getConnected();

			// remove all lines being deleted
			for (Item ic : lineEnd.getConnected()) {
				if (ic instanceof Line
						&& ((Line) ic).getOppositeEnd(lineEnd) != null) {
					Line line = (Line) ic;

					Item d = line.getOppositeEnd(lineEnd);
					d.removeLine(line);

					// if the dot was only part of one line, it can be
					// removed
					if (d.getLines().size() == 0) {
						if (d.getParent() != null)
							d.getParent().removeItem(d);
						if (!copy.contains(d))
							copy.add(d);
					}

					if (lineEnd.getParent() != null)
						lineEnd.getParent().removeItem(ic);
				}
			}
			return copy;
		}
	}

	private static void removeAllLinesExcept(Item from, Item but) {
		List<Line> lines = new LinkedList<Line>();
		lines.addAll(from.getLines());
		for (Line line : lines)
			if (line.getOppositeEnd(from) != but)
				from.removeLine(line);

		List<Constraint> consts = new LinkedList<Constraint>();
		consts.addAll(from.getConstraints());
		for (Constraint c : consts)
			if (c.getOppositeEnd(from) != but)
				from.removeConstraint(c);
	}

	public static Collection<Item> merge(List<Item> merger, Item mergee) {
		if (mergee.getParent().getNameItem() == mergee) {
			return mergee.getParent().merge(merger);
		}

		// check for rectangle merging
		if (merger.size() == 3 && mergee.getLines().size() == 2) {
			Item corner = getShapeCorner(merger);
			// if this is a corner of a shape
			if (corner != null) {
				Collection<Item> allConnected = corner.getAllConnected();
				// Check if we are collapsing a rectangle
				if (allConnected.size() == 8 && allConnected.contains(mergee)) {
					DisplayIO.setCursorPosition(mergee.getPosition());
					DisplayIO.getCurrentFrame().removeAllItems(allConnected);

					// find the point opposite corner...
					Item opposite = null;
					List<Line> lines = corner.getLines();
					for (Line l : lines) {
						allConnected.remove(l.getOppositeEnd(corner));
					}
					allConnected.remove(corner);
					for (Item i : allConnected) {
						if (i.isLineEnd()) {
							opposite = i;
							break;
						}
					}
					assert (opposite != null);

					// check if the rectangle is small enough that it should be
					// collapsed to a single point
					int x1 = Math.abs(opposite.getX() - mergee.getX());
					int x2 = Math.abs(opposite.getY() - mergee.getY());
					int distance = (int) Math.sqrt(Math.pow(x1, 2)
							+ Math.pow(x2, 2));

					if (distance < RECTANGLE_TO_POINT_THRESHOLD) {
						mergee.removeAllConstraints();
						mergee.removeAllLines();
						mergee.setThickness(4 * mergee.getThickness());
						return mergee.getAllConnected();
					} else {
						removeAllLinesExcept(mergee, opposite);
						removeAllLinesExcept(opposite, mergee);

						return mergee.getAllConnected();
					}
				}
			}
		}

		List<Item> remain = new ArrayList<Item>();
		Item res = null;

		for (Item i : merger) {
			// check for link merging
			if (i instanceof Text
					&& FrameIO.isValidFrameName((((Text) i).getFirstLine()))
					&& FrameIO.DoesFrameExist((((Text) i).getFirstLine()))) {
				// check that we can actually access the frame this link
				// corresponds to

				mergee.setLink(((Text) i).getFirstLine());
			} else {
				// check for attribute merging
				if (i instanceof Text && !i.isLineEnd()) {
					Text txt = (Text) i;

					// if this is not an attribute merge
					if (!AttributeUtils.SetAttribute(mergee, txt)) {
						// set mouse position for text merges
						if (mergee instanceof Text)
							((Text) mergee).insertText(txt.getText(), DisplayIO
									.getMouseX(), DisplayIO.getMouseY());

						else if (mergee instanceof Dot) {
							DisplayIO.setCursorPosition(mergee.getPosition());
							txt.setPosition(mergee.getPosition());
							txt.setThickness(mergee.getThickness());
							Frame parent = mergee.getParent();
							parent.removeItem(mergee);
							anchor(txt);
							// change the end points of the lines to the text
							// item
							while (mergee.getLines().size() > 0) {
								Line l = mergee.getLines().get(0);
								l.replaceEnd(mergee, txt);
							}
							break;
						}

						// TODO tidy this up...
						// Dot override doesnt use the x and y coords
						// Text does... but could be removed
						res = mergee.merge(i, DisplayIO.getMouseX(), MouseY);
						if (res != null) {
							remain.add(res);
						}
					}
				} else {
					if (mergee.isLineEnd()) {
						DisplayIO.setCursorPosition(mergee.getPosition());
					}
					// Moving the cursor ensures shapes are anchored correctly
					res = mergee.merge(i, DisplayIO.getMouseX(), MouseY);
					if (res != null)
						remain.addAll(res.getConnected());

				}
			}
		}
		updateCursor();
		mergee.getParent().setChanged(true);
		ItemUtils.EnclosedCheck(mergee.getParent().getItems());
		FrameUtils.Parse(mergee.getParent());

		return remain;
	}

	/**
	 * Picks up an item on a frame.
	 * 
	 * @param toGrab
	 *            item to be picked up
	 * @param removeItem
	 *            true if the item should be removed from the frame
	 */
	public static void pickup(Item toGrab) {
		if (toGrab.isFrameName())
			return;

		if (toGrab.Permission < Item.PERMISSION_FULL) {
			FrameGraphics
					.DisplayMessage("Insufficient permission pickup the item");
			return;
		}
		// Brook: If the widget corner is being picked up. Instead refer to
		// picking up the edge for fixed-sized widgets so it is not so confusing
		if (toGrab instanceof WidgetCorner) {
			WidgetCorner wc = (WidgetCorner) toGrab;
			if (wc.getWidgetSource().isFixedSize()) {
				for (Item i : toGrab.getConnected()) {
					if (i instanceof WidgetEdge) {
						toGrab = i;
						break;
					}
				}
			}
		}
		pickup(toGrab.getConnected());
	}

	public static void pickup(Collection<Item> toGrab) {

		String currentFrame = DisplayIO.getCurrentFrame().getName();
		Iterator<Item> iter = toGrab.iterator();
		while (iter.hasNext()) {
			Item i = iter.next();
			if (i.Permission < Item.PERMISSION_FULL) {
				iter.remove();
				continue;
			}

			// Check if it has a relative link if so make it absolute
			i.setAbsoluteLink();
			// parent may be null
			if (i.getParent() != null) {
				i.getParent().removeItem(i);
				if (currentFrame.equals(i.getParent().getName()))
					i.setParent(null);
			}
			Frame.FreeItems.add(i);
			i.setFloating(true);
		}

		updateCursor();

		// if there are multiple items in the list, determine which to use for
		// offset calculations
		if (toGrab.size() > 1) {
			for (Item i : toGrab) {
				// MIKE: Why cant we use a line for the offset calculations?!?
				if (!(i instanceof Line)) {
					_offX = DisplayIO.getMouseX() - i.getX() + i.getOffset().x;
					_offY = MouseY - i.getY() + i.getOffset().y;

					// make the offset item the first item in the list (so
					// move method knows which item to use)
					Frame.FreeItems.set(Frame.FreeItems.indexOf(i),
							Frame.FreeItems.get(0));
					Frame.FreeItems.set(0, i);
					break;
				}
			}

			move(Frame.FreeItems);
			ItemUtils.EnclosedCheck(toGrab);
			// otherwise, just use the first item
		} else if (toGrab.size() == 1) {
			Item soleItem = toGrab.iterator().next();
			_offX = DisplayIO.getMouseX() - soleItem.getX()
					+ soleItem.getOffset().x;
			_offY = MouseY - soleItem.getY() + soleItem.getOffset().y;
			// Now call move so that if we are on a message in the message box
			// It doesnt appear up the top of the scree!!
			move(toGrab);
		} else {
			FrameGraphics
					.DisplayMessage("Insufficient permission pickup the items");
		}
	}

	private static Line createLine() {
		// create the two endpoints
		Item end = DisplayIO.getCurrentFrame().createDot();
		Item start = DisplayIO.getCurrentFrame().createDot();
		// create the Line
		Line line = new Line(start, end, DisplayIO.getCurrentFrame()
				.getNextItemID());
		line.autoArrowheadLength();

		// anchor the start
		anchor(start);

		// attach the line to the cursor
		pickup(end);

		return line;
	}

	/**
	 * Returns a list of copies of the list passed in
	 * 
	 * @param toCopy
	 *            The list of items to copy
	 * @return A List of copied Items
	 */
	private static List<Item> copy(Collection<Item> toCopy) {
		return ItemUtils.CopyItems(toCopy);
	}

	public static void anchor(Item toAnchor, boolean checkEnclosure) {
		toAnchor.anchor();

		if (checkEnclosure) {
			FrameGraphics.Repaint();
			ItemUtils.EnclosedCheck(toAnchor.getParentOrCurrentFrame()
					.getItems());
		}
	}

	public static void anchor(Item toAnchor) {
		anchor(toAnchor, true);
	}

	public static void anchor(Collection<Item> toAnchor) {
		// Need to make sure we check enclosure for overlays etc
		Set<Frame> checkEnclosure = new HashSet<Frame>();
		for (Item i : toAnchor) {
			anchor(i, false);
			checkEnclosure.add(i.getParentOrCurrentFrame());
		}
		toAnchor.clear();
		FrameGraphics.Repaint();
		// Check enclosure for all the frames of the items that were anchored
		for (Frame f : checkEnclosure) {
			ItemUtils.EnclosedCheck(f.getItems());
		}
	}

	/*
	 * private boolean mouseMovedRecently() { Date now = new Date();
	 * 
	 * return now.getTime() - _lastMouseMovement.getTime() < 150; }
	 */

	public void mouseWheelMoved(MouseWheelEvent arg0) {
		NavigationActions.ResetLastAddToBack();

		int clicks = arg0.getClickCount();

		// if a line is being rubber-banded, check for auto straightening
		if (Frame.FreeItems.size() == 2) {
			if ((Frame.FreeItems.get(0) instanceof Dot && Frame.FreeItems
					.get(1) instanceof Line)
					|| (Frame.FreeItems.get(1) instanceof Dot && Frame.FreeItems
							.get(0) instanceof Line)) {

				Line line;
				if (Frame.FreeItems.get(0) instanceof Line)
					line = (Line) Frame.FreeItems.get(0);
				else
					line = (Line) Frame.FreeItems.get(1);

				// User must do multiple clicks to toggle the line
				if (clicks == MOUSE_WHEEL_THRESHOLD) {
					if (arg0.isShiftDown())
						line.toggleArrowHeadRatio(arg0.getWheelRotation());
					else
						line.toggleArrowHeadLength(arg0.getWheelRotation());
					// line.getParent().change();
					FrameGraphics.Repaint();
				}
			}
		} else if (arg0.getWheelRotation() != 0 && arg0.isShiftDown()) {

			FunctionKey rotationType = FunctionKey.SizeUp;
			if (arg0.getWheelRotation() > 0) {
				rotationType = FunctionKey.SizeDown;
			}

			Item ip = FrameUtils.getCurrentItem();
			if (ip != null && clicks > 1) {
				float size = ip.getSize();
				if (ip instanceof Dot || ip instanceof Line)
					size = ip.getThickness();
				// base the number of clicks on the size of the object
				clicks = (int) Math.ceil(size / 20.0 * clicks);
			}
			FrameKeyboardActions.functionKey(rotationType, clicks);

		} else if (clicks == MOUSE_WHEEL_THRESHOLD) {
			Item item = FrameUtils.getCurrentItem();

			// if the user is not pointing to any item
			if (item == null) {
				FrameKeyboardActions.NextTextItem(null,
						arg0.getWheelRotation() > 0);
				return;
			}

			if (item instanceof Line) {
				// check permissions
				if (item.Permission < Item.PERMISSION_FULL) {
					FrameGraphics
							.DisplayMessage("Insufficient permission to edit the Line");
					return;
				}
				Line line = (Line) item;
				line.toggleDashed(arg0.getWheelRotation());
				line.getParent().change();
				FrameGraphics.Repaint();
				return;
			} else if (item instanceof Text) {
				FrameKeyboardActions.NextTextItem(item,
						arg0.getWheelRotation() > 0);
			}
		}
	}

	/**
	 * 
	 * @return the integer value for the last mouse button clicked.
	 */
	public static int getLastMouseButton() {
		if (_lastMouseClick == null)
			return MouseEvent.NOBUTTON;

		return _lastMouseClick.getButton();
	}

	public static boolean isDelete(int modifiersEx) {

		int onMask = MouseEvent.BUTTON3_DOWN_MASK
				| MouseEvent.BUTTON2_DOWN_MASK;
		return (modifiersEx & onMask) == onMask;
	}

	public static boolean isGetAttributes(int modifiersEx) {
		int onMask = MouseEvent.BUTTON3_DOWN_MASK
				| MouseEvent.BUTTON1_DOWN_MASK;
		return (modifiersEx & onMask) == onMask;
	}

	public static boolean isTwoClickNoOp(int modifiersEx) {
		int onMask = MouseEvent.BUTTON2_DOWN_MASK
				| MouseEvent.BUTTON1_DOWN_MASK;
		return (modifiersEx & onMask) == onMask;
	}

	public static boolean wasDeleteClicked() {
		if (_lastMouseClick == null)
			return false;
		return isDelete(_lastMouseClickModifiers);
	}
}
