package org.expeditee.io;

import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.LinkedHashMap;
import java.util.List;

import org.expeditee.gui.Frame;
import org.expeditee.items.Constraint;
import org.expeditee.items.Item;
import org.expeditee.items.Line;
import org.expeditee.items.Text;

public abstract class FrameReader {
	protected static LinkedHashMap<Character, Method> _ItemTags = null;

	protected static LinkedHashMap<Character, Method> _FrameTags = null;
	
	protected static Class[] pString = { String.class };
	protected static Class[] pInt = { int.class };
	protected static Class[] pFloat = { float.class };
	protected static Class[] pColor = { Color.class };
	protected static Class[] pBool = { boolean.class };
	protected static Class[] pFont = { Font.class };
	protected static Class[] pPoint = { Point.class };
	protected static Class[] pArrow = { int.class, double.class };
	protected static Class[] pList = { List.class };
	protected static Class[] pIntArray = { int[].class };
	protected static Class[] pItem = { Item.class };
	
	public FrameReader(){
		if (_ItemTags != null && _FrameTags != null)
			return;

		_ItemTags = new LinkedHashMap<Character, Method>();
		_FrameTags = new LinkedHashMap<Character, Method>();

		try {
			_FrameTags.put('A', Frame.class.getMethod("setName", pString));
			_FrameTags.put('V', Frame.class.getMethod("setVersion", pInt));
			_FrameTags
					.put('p', Frame.class.getMethod("setProtection", pString));
			_FrameTags.put('U', Frame.class.getMethod("setOwner", pString));
			_FrameTags.put('D', Frame.class
					.getMethod("setDateCreated", pString));
			_FrameTags.put('M', Frame.class.getMethod("setLastModifyUser",
					pString));
			_FrameTags.put('d', Frame.class.getMethod("setLastModifyDate",
					pString));
			_FrameTags
					.put('F', Frame.class.getMethod("setFrozenDate", pString));

			_FrameTags.put('O', Frame.class.getMethod("setForegroundColor",
					pColor));
			_FrameTags.put('B', Frame.class.getMethod("setBackgroundColor",
					pColor));

			_ItemTags.put('S', Item.class.getMethod("setID", pInt));
			_ItemTags.put('s', Item.class.getMethod("setDateCreated", pString));
			_ItemTags.put('d', Item.class.getMethod("setColor", pColor));
			_ItemTags.put('G', Item.class.getMethod("setBackgroundColor",
					pColor));
			_ItemTags.put('P', Item.class.getMethod("setPosition", pPoint));
			_ItemTags.put('F', Item.class.getMethod("setLink", pString));
			_ItemTags.put('X', Item.class.getMethod("setActions", pList));
			_ItemTags.put('x', Item.class.getMethod("setActionMark", pBool));
			_ItemTags.put('U', Item.class.getMethod("setActionCursorEnter",
					pList));
			_ItemTags.put('V', Item.class.getMethod("setActionCursorLeave",
					pList));
			_ItemTags.put('W', Item.class.getMethod("setActionEnterFrame",
					pList));
			_ItemTags.put('Y', Item.class.getMethod("setActionLeaveFrame",
					pList));
			_ItemTags.put('D', Item.class.getMethod("setData", pList));
			_ItemTags.put('u', Item.class.getMethod("setHighlight", pBool));
			_ItemTags.put('e', Item.class.getMethod("setFillColor", pColor));
			_ItemTags.put('i', Item.class.getMethod("setFillPattern", pString));
			_ItemTags.put('o', Item.class.getMethod("setOwner", pString));
			_ItemTags.put('n', Item.class.getMethod("setLinkMark", pBool));
			_ItemTags
					.put('q', Item.class.getMethod("setLinkFrameset", pString));
			_ItemTags
					.put('y', Item.class.getMethod("setLinkTemplate", pString));
			_ItemTags.put('g', Item.class
					.getMethod("setLinePattern", pIntArray));

			_ItemTags.put('j', Item.class.getMethod("setArrow", pArrow));

			_ItemTags.put('f', Text.class.getMethod("setFont", pFont));
			_ItemTags.put('t', Text.class.getMethod("setSpacing", pInt));
			_ItemTags.put('T', Text.class.getMethod("appendLine", pString));
			_ItemTags.put('a', Text.class.getMethod("setWordSpacing", pInt));
			_ItemTags.put('b', Text.class.getMethod("setLetterSpacing", pInt));
			_ItemTags.put('m', Text.class.getMethod("setInitialSpacing", pInt));
			_ItemTags.put('w', Text.class.getMethod("setWidth", pInt));
			_ItemTags.put('k', Text.class.getMethod("setJustification", pInt));

			_ItemTags.put('h', Item.class.getMethod("setThickness", pFloat));
			_ItemTags.put('l', Item.class.getMethod("setLineIDs", pString));
			_ItemTags.put('c', Item.class
					.getMethod("setConstraintIDs", pString));
			// Lines and constraints are created differently
			_ItemTags.put('L', Line.class.getMethod("setStartItem", pItem));
			_ItemTags.put('C', Constraint.class.getMethod("getID",
					(Class[]) null));
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}
	
	public abstract Frame readFrame(String fullPath) throws IOException;
}
