package org.expeditee.io;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.List;

import org.expeditee.actions.Misc;
import org.expeditee.gui.FrameIO;
import org.expeditee.items.Item;
import org.expeditee.items.Picture;
import org.expeditee.items.Text;

public class HTMLBWriter extends AbstractHTMLWriter {

	@Override
	protected void writeTitle(Text toWrite, List<Item> items)
			throws IOException {
		String heading = toWrite.getText();
		String tag = "h" + (getIndent() + 1);
		// indenting for tag
		indent();
		_writer.write("<" + tag + ">" + heading + "</" + tag + ">"
				+ ItemWriter.NEW_LINE);
	}

	@Override
	protected void writeText(Text text) throws IOException {
		indent();
		_writer.write("<p>");
		_writer.write(text.getText().replace('\n', ' '));
		_writer.write("</p>" + ItemWriter.NEW_LINE);
	}

	@Override
	protected void writePicture(Picture pic) throws IOException {
		Image image = pic.getImage();
		String filesFolder = getFilesFolder();
		String fileName;
		// If its a bufferedImage then just write it out to the files directory
		//This means it is probably a FrameImage
		if (image instanceof BufferedImage) {
			String link = pic.getAbsoluteLink();
			// Account for the possiblitly of an unlinked buffered image
			fileName = link == null ? ("Image" + pic.getID()) : link;
			fileName = Misc.SaveImage((BufferedImage) image, "PNG", FrameIO.EXPORTS_DIR
					+ filesFolder, fileName);
		} else {//It is a normal Image stored somewhere
			Text source = pic.getText();

			fileName = source.getFirstLine();
			fileName = fileName.substring(fileName.indexOf(":") + 1).trim();

			String oldImageName = FrameIO.IMAGES_PATH + fileName;
			String newImageName = FrameIO.EXPORTS_DIR + filesFolder + fileName;
			try {
				FrameIO.copyFile(oldImageName, newImageName);
			} catch (Exception e) {
				filesFolder = "";
			}
			if (filesFolder.equals("")) {
				filesFolder = "../" + FrameIO.IMAGES_FOLDER;
			}
		}
		indent();
		StringBuffer imageTag = new StringBuffer("<img src=\"");
		imageTag.append(filesFolder).append(fileName);
		imageTag.append("\" height=").append(pic.getHeight());
		imageTag.append(" width=").append(pic.getWidth());
		imageTag.append(" border=1>");
		_writer.write(imageTag.toString());
		_writer.write(ItemWriter.NEW_LINE);
	}
}
