package org.expeditee.items;

import java.awt.image.ImageObserver;

import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.gui.FrameIO;

public class FrameImage extends FramePicture {
	/**
	 * Creates a new Picture from the given path. The ImageObserver is optional
	 * and can be set to NULL. <br>
	 * Note: It is assumed that the file described in path has already been
	 * checked to exist.
	 * 
	 * @param source
	 *            The Text Item that was used to create this Picture
	 * @param path
	 *            The Path of the Image to load from disk.
	 * @param observer
	 *            The ImageObserver to assign when painting the Image on the
	 *            screen.
	 */
	public FrameImage(Text source, String size, ImageObserver observer)
			throws IllegalArgumentException {
		super();
		_source = source;
		_imageObserver = observer;

		refresh();

		if (_image == null)
			throw new IllegalArgumentException();

		parseSize(size);

	}

	@Override
	protected Picture createPicture() {
		return ItemUtils
				.CreateFrameImage((Text) _source.copy(), _imageObserver);
	}

	@Override
	public boolean refresh() {
		assert (_source.getLink() != null);
		Frame frame = FrameIO.LoadFrame(_source.getAbsoluteLink());
		if (frame == null)
			return false;

		frame.setBuffer(null);
		FrameGraphics.UpdateBuffer(frame, false);
		_image = frame.getBuffer();

		return true;
	}
	
	@Override
	protected String getTagText() {
		return "@f: ";
	}
}
