package org.expeditee.stats;

import java.sql.Time;

import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.io.Logger;

public class CometStats {
	protected static String COLUMN_SEPARATOR = " | ";
	
	protected int _sessions = 0;

	protected String _name = null;

	protected long _active = 0;

	protected long _dark = 0;

	public CometStats(Frame topFrame) {
		this(topFrame, false);
	}

	public CometStats(Frame topFrame, boolean includeCurrentSession) {
		_name = topFrame.getName();
		Frame backupFrame = topFrame;
		if (includeCurrentSession) {
			_active = SessionStats.getFrameActiveTime().getTime();
			if (_active > 0) {
				_sessions = 1;
				_dark = SessionStats.getFrameDarkTime().getTime();
			}
		}

		// Keep going back until we have reached the end of the comet tail
		while (backupFrame != null) {
			long backupActive = backupFrame.getActiveTime().getTime();
			// Update the time if user spent time editing the frame
			// backupActive will be zero if the frame has not been saved yet
			if (backupActive > 0) {
				_sessions++;
				_active += backupActive;
				_dark += backupFrame.getDarkTime().getTime();
			}
			FrameGraphics.OverwriteMessage("Computed: " + backupFrame.getName());
			// Get the backup of the backup;
			backupFrame = backupFrame.getBackupFrame();
		}
	}

	@Override
	public String toString() {
		Time total = new Time(_active + _dark);
		Time active = new Time(_active);
		Time dark = new Time(_dark);

		StringBuffer sb = new StringBuffer();
		sb.append("CometName: ").append(_name).append('\n');
		sb.append("Edits: ").append(_sessions).append('\n');
		sb.append("            Current  | Average  | Total").append('\n');
		Time averageActive = new Time((long) (0.5 + active.getTime()
				/ (1.0 * _sessions)));
		Time averageDark = new Time((long) (0.5 + dark.getTime()
				/ (1.0 * _sessions)));
		Time averageTotal = new Time((long) (0.5 + total.getTime()
				/ (1.0 * _sessions)));
		Time currentActive = SessionStats.getFrameActiveTime();
		Time currentDark = SessionStats.getFrameDarkTime();
		Time currentTotal = SessionStats.getFrameTotalTime();

		sb.append("ActiveTime: ").append(Logger.EasyTimeFormat(currentActive))
				.append(COLUMN_SEPARATOR).append(Logger.EasyTimeFormat(averageActive))
				.append(COLUMN_SEPARATOR).append(Logger.EasyTimeFormat(active)).append('\n');
		sb.append("  DarkTime: ").append(Logger.EasyTimeFormat(currentDark))
				.append(COLUMN_SEPARATOR).append(Logger.EasyTimeFormat(averageDark)).append(
						COLUMN_SEPARATOR).append(Logger.EasyTimeFormat(dark)).append('\n');
		sb.append(" TotalTime: ").append(Logger.EasyTimeFormat(currentTotal))
				.append(COLUMN_SEPARATOR).append(Logger.EasyTimeFormat(averageTotal))
				.append(COLUMN_SEPARATOR).append(Logger.EasyTimeFormat(total));
		return sb.toString();
	}
}
